import os
import glob
import shutil
import pickle
import numpy as np
import pandas as pd
from pysagas.cfd import OPM
from pysagas.flow import FlowState
from hypervehicle.generator import Generator
from pyoptsparse import Optimizer, Optimization
from pysagas.geometry.parsers import PyMesh, STL
from pysagas.optimisation.optimiser import ShapeOpt
from typing import List, Dict, Optional, Optional, Any
from pysagas.sensitivity import GenericSensitivityCalculator
from hypervehicle.utilities import SensitivityStudy, merge_stls


class OPMShapeOpt(ShapeOpt):
    def __init__(
        self,
        optimiser: Optimizer,
        vehicle_generator: Generator,
        objective_callback: callable,
        jacobian_callback: callable,
        freestream: FlowState,
        geometry_filename: str,
        A_ref: Optional[float] = 1.0,
        optimiser_options: Optional[Dict[str, Any]] = None,
        sensitivity_filename: str = "all_components_sensitivity.csv",
        working_dir_name: str = "working_dir",
        basefiles_dir_name: str = "basefiles",
        save_evolution: bool = True,
        verbosity: int = 1,
        sensitivity_kwargs: dict = None,
        parser_kwargs: dict = None,
    ) -> None:
        # Declare global variables
        global fs_flow
        global generator
        global working_dir
        global obj_cb, jac_cb
        global save_geom, evo_dir
        global sens_filename, basefiles_dir
        global geom_filename
        global _A_ref
        global verbose

        # Dynamics global variables
        global cells, solver
        cells = None
        solver = None

        # Construct paths
        # TODO - I dont think basefiles is needed.
        home_dir = os.getcwd()
        basefiles_dir = os.path.join(home_dir, basefiles_dir_name)
        working_dir = os.path.join(home_dir, working_dir_name)
        sens_filename = sensitivity_filename
        generator = vehicle_generator
        save_geom = save_evolution
        verbose = verbosity

        # Save sensitivity kwargs
        global _sens_kwargs, _parser_kwargs
        _sens_kwargs = sensitivity_kwargs if sensitivity_kwargs else {}
        _parser_kwargs = parser_kwargs if parser_kwargs else {}

        if save_geom:
            # Create evolution history directory
            evo_dir = os.path.join(home_dir, "evolution")
            if not os.path.exists(evo_dir):
                os.mkdir(evo_dir)

        # Save callback functions
        obj_cb = objective_callback
        jac_cb = jacobian_callback

        # Construct optimisation problem
        self.opt_problem = Optimization(
            name="Cart3D-PySAGAS Shape Optimisation",
            objFun=evaluate_objective,
            sens=evaluate_gradient,
        )

        # Other
        fs_flow = freestream
        geom_filename = geometry_filename
        _A_ref = A_ref

        # Complete initialisation
        optimiser_options = optimiser_options if optimiser_options else {}
        super().__init__(optimiser, working_dir, optimiser_options)


def evaluate_objective(x: dict):
    if verbose > 0:
        print("\n\033[1mEvaluating objective function\033[0m")

    # Pre-process parameters
    _process_parameters(x)

    # Run flow solver
    loads_dict = _run_simulation()

    # Load properties
    properties_dir = glob.glob("*_properties")
    if properties_dir:
        # Load volume and mass
        volmass = pd.read_csv(
            glob.glob(os.path.join(properties_dir[0], "*volmass.csv"))[0],
            index_col=0,
        )

        # Load COG data
        cog = np.loadtxt(
            glob.glob(os.path.join(properties_dir[0], "*cog.txt"))[0], delimiter=","
        )

        # Fetch user-defined properties
        properties_file = glob.glob(os.path.join(properties_dir[0], "*properties.csv"))
        if len(properties_file) > 0:
            # File exists, load it
            properties = pd.read_csv(
                properties_file[0],
                index_col=0,
            )["0"]
        else:
            properties = None

    else:
        # No properties data found
        volmass = None
        properties = None
        cog = None

    # Evaluate objective function
    funcs = obj_cb(
        loads_dict=loads_dict,
        volmass=volmass,
        properties=properties,
        parameters=x,
        cog=cog,
    )
    failed = False

    if verbose > 0:
        print("  Done evaluating objective function.")

    return funcs, failed


def evaluate_gradient(x: dict, objective: dict):
    if verbose > 0:
        print("\n\033[1mEvaluating gradient function\033[0m")

    # Pre-process parameters
    _process_parameters(x)

    # Run flow solver for gradient information
    loads_dict, coef_sens = _run_sensitivities()

    # Calculate Jacobian
    properties_dir = glob.glob("*_properties")
    if properties_dir:
        scalar_sens_dir = os.path.join("scalar_sensitivities")
        vm = pd.read_csv(
            glob.glob(os.path.join(properties_dir[0], "*volmass.csv"))[0],
            index_col=0,
        )
        vm_sens = pd.read_csv(
            os.path.join(scalar_sens_dir, "volmass_sensitivity.csv"),
            index_col=0,
        )[x.keys()]

        # Fetch user-defined properties and sensitivities
        properties_file = glob.glob(os.path.join(properties_dir[0], "*properties.csv"))
        if len(properties_file) > 0:
            # File exists, load it
            properties = pd.read_csv(
                properties_file[0],
                index_col=0,
            )["0"]

            # Also load sensitivity file
            property_sens = pd.read_csv(
                os.path.join(scalar_sens_dir, "property_sensitivity.csv"),
                index_col=0,
            )[x.keys()]
        else:
            properties = None
            property_sens = None

    else:
        # No properties data found
        vm = None
        vm_sens = None
        properties = None
        property_sens = None

    # Call function
    jac = jac_cb(
        parameters=x,
        coef_sens=coef_sens,
        loads_dict=loads_dict,
        volmass=vm,
        volmass_sens=vm_sens,
        properties=properties,
        property_sens=property_sens,
    )

    if verbose > 0:
        print("  Done evaluating gradient function.")

    return jac


def _process_parameters(x):
    # Move into working directory
    os.chdir(working_dir)

    # Load existing parameters to compare
    already_started = _compare_parameters(x)

    if already_started and verbose > 0:
        print("  Picking up solution from file...")

    if not already_started:
        # These parameters haven't run yet, prepare the working directory
        if save_geom:
            # Move components file into evolution directory
            geom_filepath = os.path.join(working_dir, geom_filename)

            # Check if file exists
            if os.path.exists(geom_filepath):
                # Determine new name
                suffix = geom_filename.split(".")[-1]
                new_filename = f"{len(os.listdir(evo_dir)):04d}.{suffix}"

                # Copy file over
                shutil.copyfile(
                    geom_filepath,
                    os.path.join(evo_dir, new_filename),
                )

        # Delete any files from previous run
        _clean_dir(working_dir)

    # Dump design parameters to file
    with open("parameters.pkl", "wb") as f:
        pickle.dump(x, f)

    # Generate vehicle and geometry sensitivities
    if len(glob.glob("*sensitivity*")) == 0 or not already_started:
        # No sensitivity files generated yet, or this is new geometry
        if verbose > 0:
            print("  Running geometric sensitivity study.")
        parameters = _unwrap_x(x)
        ss = SensitivityStudy(vehicle_constructor=generator, verbosity=0)
        ss.dvdp(parameter_dict=parameters, perturbation=2, write_nominal_stl=True)
        ss.to_csv()

        # Merge STLs
        vehicle = ss.nominal_vehicle_instance
        stl_files = [f"{c}.stl" for c in vehicle.get_non_ghost_components().keys()]
        if len(stl_files) > 1:
            geom_file = merge_stls(
                stl_files=stl_files, name=vehicle.name, verbosity=verbose
            )
        else:
            geom_file = stl_files[0]


def _run_simulation():
    """Prepare and run the CFD simulation with the OPM solver."""
    global cells, solver

    cells = None
    solver = None

    # Load cells from geometry
    if cells is None:
        try:
            cells = PyMesh.load_from_file(
                geom_filename, verbosity=verbose, **_parser_kwargs
            )
        except:
            cells = STL.load_from_file(
                geom_filename, verbosity=verbose, **_parser_kwargs
            )

    # Run OPM solver
    # if solver is None:
    solver = OPM(cells=cells, freestream=fs_flow, verbosity=verbose)

    if verbose > 0:
        print("  Running OPM flow solver.")

    sim_results = solver.solve()

    # Construct coefficient dictionary
    CL, CD, Cm = sim_results.coefficients()
    coefficients = {"CL": CL, "CD": CD, "Cm": Cm}

    return coefficients


def _run_sensitivities():
    global cells, solver

    cells = None
    solver = None

    # Load cells from geometry
    if cells is None:
        try:
            # cells = PyMesh.load_from_file(geom_filename, verbosity=verbose, **_parser_kwargs)
            cells = PyMesh.load_from_file(
                filepath=geom_filename,
                geom_sensitivities=sens_filename,
                verbosity=verbose,
                **_parser_kwargs,
            )
        except:
            cells = STL.load_from_file(
                geom_filename, verbosity=verbose, **_parser_kwargs
            )

    # Run OPM to generate nominal aero data on cells
    flow_solver = OPM(cells=cells, freestream=fs_flow, verbosity=verbose)
    aero = flow_solver.solve(freestream=fs_flow)

    # Solve for aerodynamic sensitivities
    sens_solver = GenericSensitivityCalculator(
        cells=cells,
        sensitivity_filepath=sens_filename,
        cells_have_sens_data=True,
        verbosity=verbose,
    )
    result = sens_solver.calculate(**_sens_kwargs)
    F_sense = result.f_sens

    # Non-dimensionalise
    coef_sens = F_sense / (fs_flow.q * _A_ref)

    # # Run OPM solver for coefficients
    # if solver is None:
    #     solver = OPM(cells=cells, freestream=fs_flow, verbosity=0)

    # # TODO - how will sens combining be handled? Need to use merged STL
    # # TODO - remove hard coding below
    # if verbose > 0:
    #     print("  Running OPM sensitivity flow solver.")
    # sens_results = solver.solve_sens(sensitivity_filepath="nose_sensitivity.csv")

    # # Non-dimensionalise
    # coef_sens = sens_results.f_sens / (fs_flow.q * _A_ref)

    # Construct coefficient dictionary
    CL, CD, Cm = aero.coefficients()
    coefficients = {"CL": CL, "CD": CD, "Cm": Cm}

    return coefficients, coef_sens


def _compare_parameters(x):
    """Compares the current parameters x to the last run simulation
    parameters."""
    try:
        with open("parameters.pkl", "rb") as f:
            try:
                xp = pickle.load(f)
            except EOFError:
                return False

        # Compare to current parameters
        already_run = x == xp

    except FileNotFoundError:
        # Simulation not run yet
        already_run = False

    return already_run


def _clean_dir(directory: str, keep: list = None):
    """Deletes everything in a directory except for what is specified
    in keep."""
    global cells, solver

    all_files = os.listdir(directory)
    if keep is None:
        # Convert to empty list
        keep = []

    # Define files to remove and remove them
    rm_files = set(all_files) - set(keep)
    for f in rm_files:
        if os.path.isdir(f):
            shutil.rmtree(f)
        else:
            os.remove(f)

    # Also reset dynamic global variables
    cells = None
    solver = None


def _unwrap_x(x: dict) -> dict:
    """Unwraps an ordered dictionary."""
    unwrapped = {}
    for key, val in x.items():
        if len(val) == 1:
            unwrapped[key] = val[0]
        else:
            unwrapped[key] = val
    return unwrapped



from pysagas.cfd import OPM
from pysagas.flow import FlowState
from pysagas.geometry.parsers import PyMesh


# Load STL file
cells = PyMesh.load_from_file("wedge.stl")
A_ref = 1

# Instantiate solver
freestream = FlowState(mach=8, pressure=101e3, temperature=288)
solver = OPM(cells, freestream)

# Run solver
result = solver.solve(aoa=5)

# Can also evaluate at other flow conditions, with:
# result = solver.solve(aoa=5, Mach=6) # Modified AoA/Mach
# result = solver.solve(freestream=new_freestream) # Updated freestream

# Save results to VTK file for visualisation of solution
solver.save("wedge")

# Get aero coefficients
CL, CD, Cm = solver.flow_result.coefficients()



import pandas as pd
from stl import mesh
from tqdm import tqdm
import multiprocess as mp
import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from typing import List, Optional, Union
from pysagas.geometry import Cell, Vector
from pysagas.utilities import add_sens_data


class AbstractParser(ABC):
    """Interface for a geometry parser."""

    filetype = None

    @abstractmethod
    def __init__(self, **kwargs) -> None:
        pass

    def __repr__(self) -> str:
        return f"PySAGAS {self.filetype} parser"

    def __str__(self) -> str:
        return f"PySAGAS {self.filetype} parser"

    @property
    @abstractmethod
    def filetype(self):
        # This is a placeholder for a class variable defining the parser file type
        pass

    @abstractmethod
    def load(self) -> List[Cell]:
        """Load cells from file."""

    @classmethod
    @abstractmethod
    def load_from_file(self) -> List[Cell]:
        """Convenience method for loading cells from file."""


class Parser(AbstractParser):
    def __init__(self, filepath: str, verbosity: int = 1) -> None:
        self.filepath = filepath
        self.verbosity = verbosity

    @classmethod
    def load_from_file(
        cls,
        filepath: str,
        geom_sensitivities: Optional[Union[str, pd.DataFrame]] = None,
        verbosity: Optional[int] = 1,
        **kwargs,
    ) -> List[Cell]:
        """Convenience method for loading cells from file.

        Parameters
        ----------
        filepath : str
            The filepath to the geometry.

        geom_sensitivities : str | DataFrame, optional
            The geometry sensitivity data, to optionally add to the loaded cells. This can
            be provided as a path to the data in csv format, or directly as a Pandas
            DataFrame. The default is None.

        verbosity : int, optional
            The verbosity of the code. The defualt is 1.

        **kwargs
            Additional keyword arguments can be provided to control the sensitivity matching
            algorithm.

        See Also
        --------
        pysagas.utilities.add_sens_data
        """

        # Create parser instance
        parser = cls(filepath, verbosity)

        # Load file
        cells = parser.load()

        if geom_sensitivities:
            # Check input type
            if isinstance(geom_sensitivities, str):
                # File path provided, load into dataframe
                geom_sensitivities = pd.read_csv(geom_sensitivities)

            elif not isinstance(geom_sensitivities, pd.DataFrame):
                raise TypeError("Invalid data provided for 'geom_sensitivities'.")

            # Add sensitivity data to cells
            add_sens_data(
                cells=cells,
                data=geom_sensitivities,
                verbosity=verbosity,
                **kwargs,
            )
        return cells


class STL(Parser):
    filetype = "STL"

    def load(self) -> List[Cell]:
        # Load the STL
        mesh_obj = mesh.Mesh.from_file(self.filepath)

        cells = []
        # TODO - can face ids be inferred?
        if self.verbosity > 0:
            print("\nTranscribing cells:")
            pbar = tqdm(
                total=len(mesh_obj.vectors),
                position=0,
                leave=True,
                desc="  Cell transcription progress",
            )

        for vector_triple in mesh_obj.vectors:
            vertices = [Vector.from_coordinates(v) for v in vector_triple]
            try:
                cell = Cell.from_points(vertices)
                cells.append(cell)
            except:
                pass

            # Update progress bar
            if self.verbosity > 0:
                pbar.update(1)

        if self.verbosity > 0:
            pbar.close()
            print("Done.")

        return cells


class PyMesh(Parser):
    filetype = "PyMesh STL"

    def __init__(self, filepath: str, verbosity: int = 1) -> None:
        # Import PyMesh
        try:
            import pymesh
        except ModuleNotFoundError:
            raise Exception(
                "Could not find pymesh. Please follow the "
                + "installation instructions at "
                + "https://pymesh.readthedocs.io/en/latest/installation.html"
            )
        self._pymesh = pymesh
        super().__init__(filepath, verbosity)

    def load(self) -> List[Cell]:
        def mp_wrapper(face):
            vertices = [Vector.from_coordinates(mesh_vertices[i]) for i in face]
            try:
                cell = Cell.from_points(vertices, face_ids=face)
            except:
                cell = None
            return cell

        # Load the STL
        mesh_obj = self._pymesh.load_mesh(self.filepath)

        if self.verbosity > 0:
            print("\nTranscribing cells.")

        # Create multiprocessing pool to construct cells
        cells = []
        pool = mp.Pool()
        mesh_vertices = mesh_obj.vertices
        for result in pool.map(mp_wrapper, mesh_obj.faces):
            if result is not None:
                cells.append(result)

        if self.verbosity > 0:
            print("Done.")

        return cells


class TRI(Parser):
    filetype = ".tri"

    def load(self) -> List[Cell]:
        # Parse .tri file
        tree = ET.parse(self.filepath)
        root = tree.getroot()
        grid = root[0]
        piece = grid[0]
        points = piece[0]
        cells = piece[1]

        points_data = points[0].text
        cells_data = cells[0].text

        points_data_list = [el.split() for el in points_data.splitlines()[1:]]
        points_data_list = [[float(j) for j in i] for i in points_data_list]

        cells_data_list = [el.split() for el in cells_data.splitlines()[1:]]
        cells_data_list = [[int(j) for j in i] for i in cells_data_list]

        cells = []
        if self.verbosity > 0:
            print("\nTranscribing cells:")
            pbar = tqdm(
                total=len(cells_data_list),
                position=0,
                leave=True,
                desc="  Cell transcription progress",
            )
        for vertex_idxs in cells_data_list:
            vertices = [
                Vector.from_coordinates(points_data_list[i]) for i in vertex_idxs
            ]
            cell = Cell.from_points(vertices, face_ids=vertex_idxs)
            cells.append(cell)

            # Update progress bar
            if self.verbosity > 0:
                pbar.update(1)

        if self.verbosity > 0:
            pbar.close()
            print("Done.")

        return cells

    @classmethod
    def load_from_file(cls, filepath: str, verbosity: int = 1, **kwargs) -> List[Cell]:
        """Convenience method for loading cells from file."""
        # Create parser instance
        parser = cls(filepath, verbosity)

        # Load file
        cells = parser.load()

        return cells



import copy
import pandas as pd
from typing import Optional, Union
from abc import ABC, abstractmethod
from pysagas.cfd.solver import FlowResults, SensitivityResults


class AbstractDeck(ABC):
    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def __repr__(self) -> None:
        pass

    @abstractmethod
    def __str__(self) -> None:
        pass

    @abstractmethod
    def insert(self):
        """Insert data into the deck."""
        pass

    @abstractmethod
    def to_csv(self):
        """Write the deck to csv."""
        pass

    @abstractmethod
    def from_csv(self, **kwargs):
        """Load the deck from csv."""
        pass

    # @abstractmethod
    # def interpolate(self, **kwargs):
    #     # TODO - implement with scipy.interpolate.interpn
    #     pass


class Deck(AbstractDeck):
    TYPE = "deck"

    def __init__(
        self,
        inputs: list[str],
        columns: list[str],
        **kwargs,
    ) -> None:
        """Instantiate a new deck.

        Parameters
        ----------
        inputs : list[str]
            A list of the inputs to this aerodeck. For example, ["aoa", "mach"].

        columns : list[str]
            A list of column names to use for this deck. For example, ["CL", "CD", "Cm"].
        """
        self._deck = pd.DataFrame(columns=inputs + columns)
        self._inputs = inputs

    def __repr__(self) -> None:
        return f"{self.TYPE}"

    def __str__(self) -> None:
        return self.__repr__()

    @property
    def deck(self) -> pd.DataFrame:
        return self._deck.drop_duplicates()

    def to_csv(self, file_prefix: Optional[str] = None):
        """Save the deck to CSV file.

        Parameters
        ----------
        file_prefix : str, optional
            The CSV file name prefix. If None is provided, the deck __repr__
            will be used. The default is None.
        """
        file_prefix = file_prefix if file_prefix else self.__repr__()
        self.deck.to_csv(f"{file_prefix}.csv", index=False)


class MultiDeck(AbstractDeck):
    """Collection of multiple Deck objects."""

    def __init__(
        self,
        inputs: list[str],
        parameters: list[str],
        base_deck: Deck,
        **kwargs,
    ) -> None:
        """Instantiate a new deck collection.

        Parameters
        ----------
        inputs : list[str]
            A list of the inputs to this deck. For example, ["aoa", "mach"].

        parameters : list[str]
            A list of the parameters to this deck. For example, ["wingspan", "length"].

        base_deck : Deck
            The base deck to initalise self._decks with.
        """
        self._inputs = inputs
        self._parameters = list(parameters)
        self._decks: dict[str, Deck] = {p: copy.deepcopy(base_deck) for p in parameters}


class AeroDeck(Deck):
    """Aerodynamic coefficient deck."""

    TYPE = "aerodeck"

    def __init__(
        self,
        inputs: list[str],
        columns: list[str] = ["CL", "CD", "Cm"],
        a_ref: float = 1,
        c_ref: float = 1,
    ) -> None:
        """Instantiate a new aerodeck.

        Parameters
        ----------
        inputs : list[str]
            A list of the inputs to this aerodeck. For example, ["aoa", "mach"].

        columns : list[str], optional
            A list of column names to use for this deck. The default is ["CL", "CD", "Cm"].

        a_ref : float, optional
            The reference area. The default is 1.

        c_ref : float, optional
            The reference length. The default is 1.
        """

        # Save reference properties
        self._a_ref = a_ref
        self._c_ref = c_ref
        self._columns = columns

        # Complete instantiation
        super().__init__(inputs, columns)

    def insert(self, result: Union[FlowResults, dict[str, float]], **kwargs):
        """Insert aerodynamic coefficients into the aerodeck.

        Parameters
        ----------
        result : FlowResults | dict
            The aerodynamic coefficients to be inserted, either as a PySAGAS native
            FlowResults object, or a dictionary with keys matching the data columns
            specified on instantiation of the deck.

        See Also
        --------
        FlowResults
        """
        # Check inputs
        inputs_given = [i in kwargs for i in self._inputs]
        if not all(inputs_given):
            raise Exception(
                "Please provide all input values when inserting new result."
            )

        # Process results
        if isinstance(result, FlowResults):
            # Get coefficients
            coefficients = result.coefficients(A_ref=self._a_ref, c_ref=self._c_ref)

            # Extract data
            # TODO - use columns provided in init
            data = {
                "CL": coefficients[0],
                "CD": coefficients[1],
                "Cm": coefficients[2],
            }

        else:
            # Check keys
            data_given = [i in result for i in self._columns]
            if not all(data_given):
                raise Exception(
                    "Please provide a data point for all values when inserting new result."
                )

            # Data provided directly
            data = result

        # Add inputs
        data.update(kwargs)

        # Add to deck
        self._deck = pd.concat(
            [self._deck, pd.DataFrame(data, index=[len(self._deck)])]
        )

    @classmethod
    def from_csv(
        cls, filepath: str, inputs: list[str], a_ref: float = 1, c_ref: float = 1
    ):
        """Create an Aerodeck from a CSV file.

        Parameters
        ----------
        filepath : str
            The filepath to the csv file containing aerodeck data.

        inputs : list[str]
            A list of the inputs to this aerodeck. For example, ["aoa", "mach"].

        a_ref : float, optional
            The reference area. The default is 1.

        c_ref : float, optional
            The reference length. The default is 1.
        """

        # Read data from file
        data = pd.read_csv(filepath)

        # Extract inputs and columns
        columns = list(data.columns)
        inputs = [columns.pop(columns.index(coef)) for coef in inputs]

        # Instantiate aerodeck
        aerodeck = cls(inputs=inputs, columns=columns, a_ref=a_ref, c_ref=c_ref)
        aerodeck._deck = data
        return aerodeck


class SensDeck(MultiDeck):
    TYPE = "sensdeck"

    def __init__(
        self,
        inputs: list[str],
        parameters: list[str],
        a_ref: float = 1,
        c_ref: float = 1,
        **kwargs,
    ) -> None:
        """Instantiate a new sensitivity deck.

        This object is a collection of `AeroDeck`s, containing aerodynamic sensitivity
        information.

        Parameters
        ----------
        inputs : list[str]
            A list of the inputs to this deck. For example, ["aoa", "mach"].

        parameters : list[str]
            A list of the parameters to this deck. For example, ["wingspan", "length"].

        a_ref : float, optional
            The reference area. The default is 1.

        c_ref : float, optional
            The reference length. The default is 1.

        See Also
        --------
        AeroDeck
        """

        # Save reference properties
        self._a_ref = a_ref
        self._c_ref = c_ref

        # Create base sensitivity deck
        columns = ["dCL", "dCD", "dCm"]
        base_deck = AeroDeck(inputs, columns, a_ref, c_ref)

        # Complete instantiation
        super().__init__(
            inputs=inputs, parameters=parameters, base_deck=base_deck, **kwargs
        )

    def __repr__(self) -> None:
        return f"{self.TYPE}"

    def __str__(self) -> None:
        return self.__repr__()

    def insert(self, result: SensitivityResults, **kwargs):
        # Get coefficients
        f_sens, m_sens = result.coefficients(A_ref=self._a_ref, c_ref=self._c_ref)

        for param in self._parameters:
            # Extract data for this parameter
            data = {
                "dCL": f_sens.loc[param]["dCL"],
                "dCD": f_sens.loc[param]["dCD"],
                "dCm": m_sens.loc[param]["dMz/dp"],
            }

            # Insert this data into the respective parameter deck
            self._decks[param].insert(result=data, **kwargs)

    @classmethod
    def from_csv(
        cls,
        param_filepaths: dict[str, str],
        inputs: list[str],
        a_ref: float = 1,
        c_ref: float = 1,
    ):
        """Create a SensDeck from a collection of CSV files.

        Parameters
        ----------
        param_filepaths : dict[str, str]
            A dictionary of filepaths, keyed by the associated parameter.

        inputs : list[str]
            A list of the inputs to this aerodeck. For example, ["aoa", "mach"].

        a_ref : float, optional
            The reference area. The default is 1.

        c_ref : float, optional
            The reference length. The default is 1.
        """
        decks = {}
        for param, filepath in param_filepaths.items():
            # Load data
            deck = AeroDeck.from_csv(
                filepath=filepath, inputs=inputs, a_ref=a_ref, c_ref=c_ref
            )
            decks[param] = deck

        # Extract inputs
        inputs = deck._inputs

        # Instantiate sensdeck
        sensdeck = cls(
            inputs=inputs, parameters=param_filepaths.keys(), a_ref=a_ref, c_ref=c_ref
        )

        # Overwrite with loaded decks
        sensdeck._decks = decks

        return sensdeck

    def to_csv(self, file_prefix: Optional[str] = None):
        """Save the decks to CSV file.

        Parameters
        ----------
        file_prefix : str, optional
            The CSV file name prefix. If None is provided, the deck __repr__
            will be used. The default is None.
        """
        file_prefix = file_prefix if file_prefix else self.__repr__()
        for p, deck in self._decks.items():
            deck.to_csv(file_prefix=f"{p}_{file_prefix}")

    @property
    def decks(self) -> dict[str, AeroDeck]:
        decks = {p: deck.deck for p, deck in self._decks.items()}
        return decks



from __future__ import annotations
import numpy as np
import pysagas.flow
from .vector import Vector
from numpy.typing import ArrayLike
from typing import Union, List, Optional


class Cell:
    """A triangular cell object.

    Attributes
    -----------
    p0 : Vector
        The first vertex of the cell.

    p1 : Vector
        The second vertex of the cell.

    p2 : Vector
        The third vertex of the cell.

    A : float
        The cell face area.

    n : Vector
        The cell normal.

    c : Vector
        The cell centroid.

    dndv : np.array
        The sensitivity of the cell's normal vector to each vertex.

    dAdv : np.array
        The sensitivity of the cell's area to each vertex.

    dvdp : np.array
        The sensitivity of the cell's vertices to each geometric parameter.

    dndp : np.array
        The sensitivity of the cell's normal vector to each geometric parameter.

    dAdp : np.array
        The sensitivity of the cell's area to each geometric parameter.

    dcdp : np.array
        The sensitivity of the centroid to each geometric parameter.

    flowstate : FlowState
        The flowstate associated with the cell.

    sensitivities : np.array
        An array containing the [x,y,z] force sensitivities of the cell.
    """

    def __init__(
        self, p0: Vector, p1: Vector, p2: Vector, face_ids: Optional[list[int]] = None
    ):
        """Constructs a cell, defined by three points.

        Parameters
        ----------
        p0 : Vector
            The first point defining the cell.

        p1 : Vector
            The second point defining the cell.

        p2 : Vector
            The third point defining the cell.
        """
        # Save points
        self.p0 = p0
        self.p1 = p1
        self.p2 = p2

        # Save connectivity information (PyMesh)
        self._face_ids = face_ids

        # Calculate cell properites
        self.n = -self.calc_normal(p0, p1, p2)
        self.A = self.calc_area(p0, p1, p2)
        self.c = self.calc_centroid(p0, p1, p2)

        # Calculate geometric sensitivities
        self._dndv: ArrayLike = None
        self._dAdv: ArrayLike = None
        self._dcdv: ArrayLike = None

        # Parameter sensitivities
        self.dvdp: ArrayLike = None  # vertex-parameter sensitivities
        self.dndp: ArrayLike = None  # normal-parameter sensitivities
        self.dAdp: ArrayLike = None  # area-parameter sensitivities
        self.dcdp: ArrayLike = None  # centroid-parameter sensitivities

        # FlowState
        self.flowstate: pysagas.flow.FlowState = None

        # Sensitivities
        self.sensitivities = None
        self.moment_sensitivities = None

        # Cell attributes
        self.attributes = {}

    @property
    def dndv(self):
        if self._dndv is not None:
            # Already exists, return it
            return self._dndv
        else:
            # Calculate and return it
            self._dndv = self.n_sensitivity(self.p0, self.p1, self.p2)
            return self._dndv

    @property
    def dAdv(self):
        if self._dAdv is not None:
            # Already exists, return it
            return self._dAdv
        else:
            # Calculate and return it
            self._dAdv = self.A_sensitivity(self.p0, self.p1, self.p2)
            return self._dAdv

    @property
    def dcdv(self):
        if self._dcdv is not None:
            # Already exists, return it
            return self._dcdv
        else:
            # Calculate and return it
            self._dcdv = self.c_sensitivity(self.p0, self.p1, self.p2)
            return self._dcdv

    @property
    def vertices(self):
        """The cell vertices."""
        return np.array([getattr(self, p).vec for p in ["p0", "p1", "p2"]])

    def to_dict(self):
        """Returns the Cell as a dictionary."""
        # TODO - need to think about representing the cells
        # vs. the points
        # The values at the points will need to be averaged
        # across cells
        pass

    @classmethod
    def from_points(
        cls, points: Union[List[Vector], np.array[Vector]], **kwargs
    ) -> Cell:
        """Constructs a Vector object from an array of coordinates.

        Parameters
        ----------
        points : Union[List[Vector], np.array[Vector]]
            The points defining the cell.

        Returns
        -------
        Cell
        """
        return cls(*points, **kwargs)

    def __repr__(self) -> str:
        return f"Cell({self.p0}, {self.p1}, {self.p2})"

    def __str__(self) -> str:
        return "A Cell"

    def _add_sensitivities(self, dvdp: np.array) -> None:
        """Adds the cell sensitivities to the cell.

        Parameters
        -----------
        dvdp : array
            The sensitivity of each vertex x, y and z, with respect to each
            parameter. The dimensions of dvdp are (9, p), for p parameters.
            Each 3 rows corresponds to the x, y and z sensitivities for each
            vertex of the cell. Each column is for the relevant parameter.
        """
        self.dvdp = dvdp
        self.dndp = np.dot(self.dndv, self.dvdp)
        self.dAdp = np.dot(self.dAdv, self.dvdp)
        self.dcdp = np.dot(self.dcdv, self.dvdp)

    @staticmethod
    def calc_normal(p0: Vector, p1: Vector, p2: Vector) -> Vector:
        """Calculates the normal vector of a cell defined by three points.

        Parameters
        ----------
        p0 : Vector
            The first point defining the cell.

        p1 : Vector
            The second point defining the cell.

        p2 : Vector
            The third point defining the cell.

        Returns
        --------
        normal : Vector
            The unit normal vector of the cell defined by the points.

        References
        -----------
        https://www.khronos.org/opengl/wiki/Calculating_a_Surface_Normal
        """
        # Define cell edges
        A = p1 - p0
        B = p2 - p0

        # Calculate normal components
        Nx = A.y * B.z - A.z * B.y
        Ny = A.z * B.x - A.x * B.z
        Nz = A.x * B.y - A.y * B.x

        # Construct normal vector
        normal = Vector(x=Nx, y=Ny, z=Nz)

        if normal.norm == 0:
            # Degenerate cell
            raise DegenerateCell()

        # Convert to unit vector
        unit_normal = normal / np.linalg.norm(normal.vec)

        return unit_normal

    @staticmethod
    def calc_area(p0: Vector, p1: Vector, p2: Vector) -> float:
        """Calculates the area of a cell defined by three points.

        Parameters
        ----------
        p0 : Vector
            The first point defining the cell.

        p1 : Vector
            The second point defining the cell.

        p2 : Vector
            The third point defining the cell.

        Returns
        --------
        area : float
            The area of the cell defined by the points.

        References
        -----------
        https://en.wikipedia.org/wiki/Cross_product
        """
        # Define cell edges
        A = p1 - p0
        B = p2 - p0

        # Calculate area
        S = 0.5 * np.linalg.norm(np.cross(A.vec, B.vec))
        return S

    @staticmethod
    def calc_centroid(p0: Vector, p1: Vector, p2: Vector) -> Vector:
        """Calculates the centroid of a cell defined by three points.

        Parameters
        ----------
        p0 : Vector
            The first point defining the cell.

        p1 : Vector
            The second point defining the cell.

        p2 : Vector
            The third point defining the cell.

        Returns
        --------
        c : Vector
            The centroid of the cell defined by the points.

        References
        -----------
        https://en.wikipedia.org/wiki/Centroid
        """
        cx = (p0.x + p1.x + p2.x) / 3
        cy = (p0.y + p1.y + p2.y) / 3
        cz = (p0.z + p1.z + p2.z) / 3
        return Vector(cx, cy, cz)

    @staticmethod
    def n_sensitivity(p0: Vector, p1: Vector, p2: Vector) -> np.array:
        """Calculates the sensitivity of a cell's normal vector
        to the points defining the cell analytically.

        Parameters
        ----------
        p0 : Vector
            The first point defining the cell.

        p1 : Vector
            The second point defining the cell.

        p2 : Vector
            The third point defining the cell.

        Returns
        -------
        sensitivity : np.array
            The sensitivity matrix with size m x n x p. Rows m refer to
            the vertices, columns n refer to the vertex coordinates, and
            slices p refer to the components of the normal vector.
        """
        g = (
            ((p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)) ** 2
            + ((p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)) ** 2
            + ((p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)) ** 2
        )

        # n = np.empty(3)
        # A = p0
        # B = p1
        # C = p2
        # n[0] = ((B.y-A.y)*(C.z-A.z) - (B.z-A.z)*(C.y-A.y)) / g**0.5
        # n[1] = ((B.z-A.z)*(C.x-A.x) - (B.x-A.x)*(C.z-A.z)) / g**0.5
        # n[2] = ((B.x-A.x)*(C.y-A.y) - (B.y-A.y)*(C.x-A.x)) / g**0.5

        M_sense = np.empty((3, 9))
        for row in range(3):
            if row == 0:
                f = (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)  # f_x
            elif row == 1:
                f = (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)  # f_y
            elif row == 2:
                f = (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)  # f_z

            for col in range(9):
                if col == 0:  # d/dp0.x
                    g_dash = 2 * (
                        (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                    ) * (-(p1.z - p0.z) + (p2.z - p0.z)) + 2 * (
                        (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                    ) * (
                        -(p2.y - p0.y) + (p1.y - p0.y)
                    )
                elif col == 1:  # d/dp0.y
                    g_dash = 2 * (
                        (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                    ) * (-(p2.z - p0.z) + (p1.z - p0.z)) + 2 * (
                        (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                    ) * (
                        -(p1.x - p0.x) + (p2.x - p0.x)
                    )
                elif col == 2:  # d/dp0.z
                    g_dash = 2 * (
                        (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                    ) * (-(p1.y - p0.y) + (p2.y - p0.y)) + 2 * (
                        (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                    ) * (
                        -(p2.x - p0.x) + (p1.x - p0.x)
                    )
                elif col == 3:  # d/dp1.x
                    g_dash = 2 * (
                        (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                    ) * (-(p2.z - p0.z)) + 2 * (
                        (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                    ) * (
                        (p2.y - p0.y)
                    )
                elif col == 4:  # d/dp1.y
                    g_dash = 2 * (
                        (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                    ) * ((p2.z - p0.z)) + 2 * (
                        (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                    ) * (
                        -(p2.x - p0.x)
                    )
                elif col == 5:  # d/dp1.z
                    g_dash = 2 * (
                        (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                    ) * (-(p2.y - p0.y)) + 2 * (
                        (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                    ) * (
                        (p2.x - p0.x)
                    )
                elif col == 6:  # d/dp2.x
                    g_dash = 2 * (
                        (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                    ) * ((p1.z - p0.z)) + 2 * (
                        (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                    ) * (
                        -(p1.y - p0.y)
                    )
                elif col == 7:  # d/dp2.y
                    g_dash = 2 * (
                        (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                    ) * (-(p1.z - p0.z)) + 2 * (
                        (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                    ) * (
                        (p1.x - p0.x)
                    )
                elif col == 8:  # d/dp2.z
                    g_dash = 2 * (
                        (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                    ) * ((p1.y - p0.y)) + 2 * (
                        (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                    ) * (
                        -(p1.x - p0.x)
                    )

                if row == 0:  # for n..x
                    # ((p1.y-p0.y)*(p2.z-p0.z) - (p1.z-p0.z)*(p2.y-p0.y))  # f_x
                    if col == 1:  # d/dp0.y
                        f_dash = -(p2.z - p0.z) + (p1.z - p0.z)
                    elif col == 2:  # d/dp0.z
                        f_dash = -(p1.y - p0.y) + (p2.y - p0.y)
                    elif col == 4:  # d/dp1.y
                        f_dash = p2.z - p0.z
                    elif col == 5:  # d/dp1.z
                        f_dash = -(p2.y - p0.y)
                    elif col == 7:  # d/dp2.y
                        f_dash = -(p1.z - p0.z)
                    elif col == 8:  # d/dp2.z
                        f_dash = p1.y - p0.y
                    else:
                        f_dash = 0
                if row == 1:  # for n.y  # CHECK
                    # f = ((p1.z-p0.z)*(p2.x-p0.x) - (p1.x-p0.x)*(p2.z-p0.z))  # f_y
                    if col == 0:  # d/dp0.x
                        f_dash = -(p1.z - p0.z) + (p2.z - p0.z)
                    elif col == 2:  # d/dp0.z
                        f_dash = -(p2.x - p0.x) + (p1.x - p0.x)  # CHECK
                    elif col == 3:  # d/dp1.x
                        f_dash = -(p2.z - p0.z)
                    elif col == 5:  # d/dp1.z
                        f_dash = p2.x - p0.x
                    elif col == 6:  # d/dp2.x
                        f_dash = p1.z - p0.z
                    elif col == 8:  # d/dp2.z
                        f_dash = -(p1.x - p0.x)
                    else:
                        f_dash = 0
                if row == 2:  # for n.z  # CHECK
                    # f = ((p1.x-p0.x)*(p2.y-p0.y) - (p1.y-p0.y)*(p2.x-p0.x))  # f_z
                    if col == 0:  # d/dp0.x
                        f_dash = -(p2.y - p0.y) + (p1.y - p0.y)
                    elif col == 1:  # d/dp0.y
                        f_dash = -(p1.x - p0.x) + (p2.x - p0.x)
                    elif col == 3:  # d/dp1.x
                        f_dash = p2.y - p0.y
                    elif col == 4:  # d/dp1.y
                        f_dash = -(p2.x - p0.x)
                    elif col == 6:  # d/dp2.x
                        f_dash = -(p1.y - p0.y)
                    elif col == 7:  # d/dp2.y
                        f_dash = p1.x - p0.x
                    else:
                        f_dash = 0

                # Assign sensitivity
                M_sense[row, col] = (
                    f_dash * (g ** (-0.5)) + f * (-1 / 2) * g ** (-3 / 2) * g_dash
                )
        return M_sense

    @staticmethod
    def A_sensitivity(p0: Vector, p1: Vector, p2: Vector) -> np.array:
        """Calculates the sensitivity of a cell's area to the points
        defining the cell analytically.

        Parameters
        ----------
        p0 : Vector
            The first point defining the cell.

        p1 : Vector
            The second point defining the cell.

        p2 : Vector
            The third point defining the cell.

        Returns
        -------
        sensitivity : np.array
            The sensitivity matrix with size m x n. Rows m refer to
            the vertices, columns n refer to the vertex coordinates.
        """
        g = (
            ((p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)) ** 2
            + ((p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)) ** 2
            + ((p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)) ** 2
        )

        A_sense = np.empty(9)
        for col in range(9):
            if col == 0:  # d/dp0.x
                g_dash = 2 * (
                    (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                ) * (-(p1.z - p0.z) + (p2.z - p0.z)) + 2 * (
                    (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                ) * (
                    -(p2.y - p0.y) + (p1.y - p0.y)
                )
            elif col == 1:  # d/dp0.y
                g_dash = 2 * (
                    (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                ) * (-(p2.z - p0.z) + (p1.z - p0.z)) + 2 * (
                    (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                ) * (
                    -(p1.x - p0.x) + (p2.x - p0.x)
                )
            elif col == 2:  # d/dp0.z
                g_dash = 2 * (
                    (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                ) * (-(p1.y - p0.y) + (p2.y - p0.y)) + 2 * (
                    (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                ) * (
                    -(p2.x - p0.x) + (p1.x - p0.x)
                )
            elif col == 3:  # d/dp1.x
                g_dash = 2 * (
                    (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                ) * (-(p2.z - p0.z)) + 2 * (
                    (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                ) * (
                    (p2.y - p0.y)
                )
            elif col == 4:  # d/dp1.y
                g_dash = 2 * (
                    (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                ) * ((p2.z - p0.z)) + 2 * (
                    (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                ) * (
                    -(p2.x - p0.x)
                )
            elif col == 5:  # d/dp1.z
                g_dash = 2 * (
                    (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                ) * (-(p2.y - p0.y)) + 2 * (
                    (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                ) * (
                    (p2.x - p0.x)
                )
            elif col == 6:  # d/dp2.x
                g_dash = 2 * (
                    (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                ) * ((p1.z - p0.z)) + 2 * (
                    (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                ) * (
                    -(p1.y - p0.y)
                )
            elif col == 7:  # d/dp2.y
                g_dash = 2 * (
                    (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                ) * (-(p1.z - p0.z)) + 2 * (
                    (p1.x - p0.x) * (p2.y - p0.y) - (p1.y - p0.y) * (p2.x - p0.x)
                ) * (
                    (p1.x - p0.x)
                )
            elif col == 8:  # d/dp2.z
                g_dash = 2 * (
                    (p1.y - p0.y) * (p2.z - p0.z) - (p1.z - p0.z) * (p2.y - p0.y)
                ) * ((p1.y - p0.y)) + 2 * (
                    (p1.z - p0.z) * (p2.x - p0.x) - (p1.x - p0.x) * (p2.z - p0.z)
                ) * (
                    -(p1.x - p0.x)
                )

            A_sense[col] = 0.5 * (1 / 2) * g ** (-1 / 2) * g_dash

        return A_sense

    @staticmethod
    def c_sensitivity(p0: Vector, p1: Vector, p2: Vector) -> np.array:
        """Calculates the sensitivity of a cell's centroid to the
        points defining the cell.

        Parameters
        ----------
        p0 : Vector
            The first point defining the cell.

        p1 : Vector
            The second point defining the cell.

        p2 : Vector
            The third point defining the cell.

        Returns
        -------
        sensitivity : np.array
            The sensitivity matrix with size m x n x p. Rows m refer to
            the vertices, columns n refer to the vertex coordinates, and
            slices p refer to the components of the centroid point.
        """
        c_sens = np.array(
            [
                [1 / 3, 0, 0, 1 / 3, 0, 0, 1 / 3, 0, 0],
                [0, 1 / 3, 0, 0, 1 / 3, 0, 0, 1 / 3, 0],
                [0, 0, 1 / 3, 0, 0, 1 / 3, 0, 0, 1 / 3],
            ]
        )
        return c_sens


class DegenerateCell(Exception):
    """Exception raised for degenerate cells."""

    def __init__(self, message="Degenerate cell"):
        self.message = message
        super().__init__(self.message)



import os
import time
import shutil
from random import random
from typing import Optional
from hypervehicle.utilities import append_sensitivities_to_tri


class C3DPrep:
    def __init__(
        self,
        logfile,
        jitter_denom: float = 1000,
        rotation_attempts: int = 6,
        info_file: str = None,
        write_config: bool = True,
    ) -> None:
        self._logfile = logfile
        self._info = info_file if info_file is not None else self._logfile
        self._jitter_denom = jitter_denom  # for 1000; Max of 0.0001, min of 0
        self._rotation_attempts = rotation_attempts
        self._write_config = write_config

    def _run_stl2tri(self, stl_files: list):
        tri_files = []
        for file in stl_files:
            prefix = file.split(".")[0]
            tri_file = prefix + ".tri"
            os.system(f"stl2tri.pl {file} {tri_file} >> {self._logfile} 2>&1")
            tri_files.append(tri_file)

        os.system(f"rm *.comp.tri *.off >> {self._logfile} 2>&1")
        return tri_files

    def _jitter_tri_files(self, tri_files):
        for file in tri_files:
            prefix = file.split(".")[0]
            x_pert = random() / self._jitter_denom
            y_pert = random() / self._jitter_denom
            z_pert = random() / self._jitter_denom
            os.system(
                f"trix -x {x_pert} -y {y_pert} -z {z_pert} -o {prefix} {file} >> {self._logfile} 2>&1"
            )

    def _shift_all(
        self,
        tri_files,
        x_shift,
        y_shift,
        z_shift,
        component: str = None,
        reverse: bool = False,
    ):
        if component is None:
            transform_files = tri_files
        else:
            transform_files = [component]

        for file in transform_files:
            prefix = ".".join(file.split(".")[:-1])
            if reverse:
                os.system(
                    f"trix -x {-x_shift} -y {-y_shift} -z {-z_shift} -o {prefix} {file} >> {self._logfile} 2>&1"
                )
            else:
                os.system(
                    f"trix -x {x_shift} -y {y_shift} -z {z_shift} -o {prefix} {file} >> {self._logfile} 2>&1"
                )

    def _rotate_all(
        self,
        tri_files,
        x_rot,
        y_rot,
        z_rot,
        component: str = None,
        reverse: bool = False,
    ):
        # Define rotations dict
        rotations = {
            "x": x_rot,
            "y": y_rot,
            "z": z_rot,
        }

        # Determine files to be transformed
        if component is None:
            # Rotate all .tri files (as provided)
            transform_files = tri_files
        else:
            # Rotate the single component
            transform_files = [component]

        for file in transform_files:
            prefix = ".".join(file.split(".")[:-1])

            # Check order of operations
            if reverse:
                order = ["z", "y", "x"]
            else:
                order = ["x", "y", "z"]

            # Apply rotations
            for axis in order:
                rotation = rotations[axis]
                os.system(
                    f"trix -r{axis} {rotation} -o {prefix} {file} >> {self._logfile} 2>&1"
                )

    def _run_comp2tri(self, tri_files):
        tri_files_str = " ".join(tri_files)

        if os.path.exists("Config.xml"):
            # Remove old config file
            os.remove("Config.xml")

        # Run command
        conf_str = "-config" if self._write_config else ""
        os.system(
            f"comp2tri -makeGMPtags {tri_files_str} {conf_str} >> {self._logfile} 2>&1"
        )

        if self._write_config:
            # Await Config.xml
            while not os.path.exists("Config.xml"):
                time.sleep(0.2)

            # Overwrite Config.xml Component names using tri files
            original = os.path.join("Config.xml")
            new = os.path.join("temp_config.xml")
            with open(new, "w+") as new_file:
                with open(original, "r") as original_file:
                    for line in original_file:
                        if "Component Name" in line:
                            # Get component number
                            name_prefix = "Component_"
                            name_start = line.index(name_prefix)
                            comp_no = line.split('"')[1].split("_")[-1]
                            tri_prefix = tri_files[int(comp_no) - 1].split(".")[0]

                            # Update line
                            line = (
                                line[:name_start]
                                + tri_prefix
                                + line[name_start + len(name_prefix) + 1 :]
                            )

                        # Write line to new file
                        new_file.write(line)

            # # Replace original aero.csh file with updated file
            shutil.copymode(original, new)
            os.remove(original)
            os.rename(new, original)

        else:
            # Await Components.tri
            while not os.path.exists("Components.tri"):
                time.sleep(0.2)

    def _run_intersect(self):
        os.system(f"intersect >> {self._logfile} 2>&1")

    @staticmethod
    def _get_stl_files() -> list[str]:
        path = os.getcwd()
        all_files = [
            f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))
        ]
        stl_files = []
        for file in all_files:
            if file.split(".")[-1] == "stl":
                stl_files.append(file)

        # Sort files
        stl_files.sort()

        return stl_files

    @staticmethod
    def _check_for_success():
        success = False
        if os.path.isfile("Components.i.tri"):
            success = True
        return success

    def intersect_stls(self, stl_files: Optional[list[str]] = None) -> bool:
        """Create Components.i.tri by intersecting all STL files."""
        # Check for existing intersected file
        if self._check_for_success():
            self._log("Intersected components file already present.")
            return True

        # Continue
        if not stl_files:
            stl_files = self._get_stl_files()
        tri_files = self._run_stl2tri(stl_files)

        # First try intersect original files
        self._log("Making first attempt to intersect components with no perturbations.")
        self._run_comp2tri(tri_files)
        self._run_intersect()
        successful = self._check_for_success()
        if successful:
            self._log("Success.")
            return True

        # That failed, try jittering components
        self._log("Attempt failed, now attempting to intersect jittered components.")
        self._jitter_tri_files(tri_files)
        self._run_comp2tri(tri_files)
        self._run_intersect()
        successful = self._check_for_success()
        if successful:
            self._log("Success.")
            return True

        # That failed, try arbitrary shifts away
        self._log("Attempt failed, now attempting random perturbations.")
        for attempt in range(self._rotation_attempts):
            # Define shifts
            x_shift = random() * 10  # Max of 10, min of 0
            y_shift = random() * 10  # Max of 10, min of 0
            z_shift = random() * 10  # Max of 10, min of 0

            # Define rotations
            x_rot = random() * 10  # Max of 10, min of 0
            y_rot = random() * 10  # Max of 10, min of 0
            z_rot = random() * 10  # Max of 10, min of 0

            # Apply transformations
            self._shift_all(
                tri_files=tri_files, x_shift=x_shift, y_shift=y_shift, z_shift=z_shift
            )
            self._rotate_all(tri_files=tri_files, x_rot=x_rot, y_rot=y_rot, z_rot=z_rot)

            if attempt > 0:
                # Also jitter
                self._log(f"On attempt {attempt+1}, also applying random jitter.")
                self._jitter_tri_files(tri_files)

            # Make intersect attempt
            self._run_comp2tri(tri_files)
            self._run_intersect()
            successful = self._check_for_success()

            if successful:
                # Move configuration back to original location
                self._log(
                    "Success. Now moving Components.i.tri back to original position."
                )
                self._rotate_all(
                    tri_files=tri_files,
                    x_rot=-x_rot,
                    y_rot=-y_rot,
                    z_rot=-z_rot,
                    component="Components.i.tri",
                    reverse=True,  # Perform in reverse order
                )
                self._shift_all(
                    tri_files=tri_files,
                    x_shift=x_shift,
                    y_shift=y_shift,
                    z_shift=z_shift,
                    component="Components.i.tri",
                    reverse=True,
                )
                return True

            else:
                # Need to reset tri files
                self._log(
                    f"Random perturbation attempt {attempt+1} failed. Resetting "
                    + ".tri files and attempting again."
                )
                tri_files = self._run_stl2tri(stl_files)

        # Finish log
        self._log("Unsuccessful.")

        return False

    def run_autoinputs(self, r: int = 2, *args):
        """Runs autoInputs to create input.c3d.

        Parameters
        -----------
        r : int, optional
            The distance to farfield normalized by max geometry size. The default is 2.
        args : str, optional
            Any extra arguments to provide to autoInputs, as strings. For arguments with
            values, just include the value in the string.
        """
        extra_args = "-" + " -".join(args) if args else ""
        os.system(f"autoInputs -r 2 {extra_args} >> {self._logfile} 2>&1")

    def run_aero(self):
        """Runs aero.csh."""
        os.system("./aero.csh restart")

    def _log(self, msg: str):
        with open(self._info, "a") as f:
            f.write("\nC3DPrep: " + msg)


def combine_sense_data(
    components_filepath: str,
    sensitivity_files: list[str],
    match_target: float = 0.9,
    tol_0: float = 1e-5,
    max_tol: float = 1e-1,
    outdir: Optional[str] = None,
    verbosity: Optional[int] = 1,
):
    """Combine the individual component sensitivity data with the
    intersected geometry file (eg. Components.i.tri).
    """
    match_frac = 0
    tol = tol_0
    while match_frac < match_target:
        # Check tolerance
        if tol > max_tol:
            raise Exception(
                "Cannot combine sensitivity data (match fraction: "
                + f"{match_frac}, tolerance: {tol}, max tolerance: {max_tol})."
            )

        # Run matching algorithm
        match_frac = append_sensitivities_to_tri(
            dp_filenames=sensitivity_files,
            components_filepath=components_filepath,
            match_tolerance=tol,
            verbosity=verbosity,
            outdir=outdir,
        )

        if match_frac < match_target and verbosity > 0:
            print(
                "Failed to combine sensitivity data "
                f"({100*match_frac:.02f}% match rate)."
            )
            print(f"  Increasing matching tolerance to {tol*10} and trying again.")
        elif verbosity > 0:
            print(
                "Component sensitivity data matched to intersected geometry "
                + f"with {100*match_frac:.02f}% match rate."
            )

        # Increase matching tolerance
        tol *= 10



__version__ = "0.12.1"

from .geometry import Vector, Cell
from .flow import GasState, FlowState


from art import tprint


def banner():
    """Prints the PySAGAS banner"""
    tprint("PySAGAS", "tarty4")



import numpy as np
from pysagas.geometry import Cell


def piston_sensitivity(cell: Cell, p_i: int, **kwargs):
    """Calculates the pressure-parameter sensitivity using
    local piston theory.

    Parameters
    ----------
    cell : Cell
        The cell object.

    p_i : int
        The index of the parameter to find the sensitivity for. This is used to
        index cell.dndp.
    """
    M_l = cell.flowstate.M
    if M_l < 1.0:
        # Subsonic cell, skip
        return 0

    dPdp = (
        cell.flowstate.rho
        * cell.flowstate.a
        * np.dot(cell.flowstate.vec, -cell.dndp[:, p_i])
    )
    return dPdp


def van_dyke_sensitivity(
    cell: Cell,
    p_i,
    **kwargs,
):
    """
    Calculates the pressure-parameter sensitivity using
    Van Dyke second-order theory.

     Parameters
    ----------
    cell : Cell
        The cell object.

    p_i : int
        The index of the parameter to find the sensitivity for. This is used to
        index cell.dndp.
    """
    piston = piston_sensitivity(cell=cell, p_i=p_i)
    M_l = cell.flowstate.M
    dPdp = piston * M_l / (M_l**2 - 1) ** 0.5
    return dPdp


def isentropic_sensitivity(cell: Cell, p_i: int, **kwargs):
    """Calculates the pressure-parameter sensitivity using
    the isentropic flow relation directly."""
    gamma = cell.flowstate.gamma
    power = (gamma + 1) / (gamma - 1)
    dPdW = (cell.flowstate.P * gamma / cell.flowstate.a) * (
        1 + cell.flowstate.v * (gamma - 1) / (2 * cell.flowstate.a)
    ) ** power
    dWdn = -cell.flowstate.vec
    dndp = cell.dndp[:, p_i]
    dPdp = dPdW * np.dot(dWdn, dndp)
    return dPdp



import numpy as np
import pandas as pd
from tqdm import tqdm
from typing import List, Optional
from pysagas.flow import FlowState
from pysagas.sensitivity.calculator import SensitivityCalculator
from pysagas.geometry import Vector, Cell, DegenerateCell
from pysagas.sensitivity.cart3d.utilities import process_components_file


class Cart3DSensitivityCalculator(SensitivityCalculator):
    """PySAGAS Cart3D flow sensitivity calculator."""

    solver = "Cart3D"

    def __init__(
        self,
        freestream: FlowState,
        sensitivity_filepath: str,
        components_filepath: Optional[str] = None,
        pointdata: Optional[pd.DataFrame] = None,
        celldata: Optional[pd.DataFrame] = None,
        write_data: Optional[bool] = False,
        verbosity: Optional[int] = 1,
        **kwargs,
    ) -> None:
        """A PySAGAS sensitivity calculator for Cart3D.

        Parameters
        ----------
        freestream : FlowState
            The flow state of the freestream.

        sensitivity_filepath : str
            The filepath to the geometry sensitivities.

        components_filepath : str, optional
            The filepath to the Components.i.plt file to be processed.
            The default is None.

        pointdata : pd.DataFrame, optional
            The point data. Must be supplied with celldata.

        celldata : pd.DataFrame, optional
            The cell data. Must be supplied with pointdata.

        write_data : bool, optional
            Write the flow data to CSV files. The default is True.

        verbosity : int, optional
            The verbosity of the code. The defualt is 1.
        """
        # Load data
        self.sensdata = pd.read_csv(sensitivity_filepath)
        if pointdata is not None and celldata is not None:
            self.pointdata = pointdata
            self.celldata = celldata
        else:
            self.pointdata, self.celldata = process_components_file(
                freestream.a,
                freestream.rho,
                components_filepath,
                write_data=write_data,
                verbosity=verbosity,
            )
        self.verbosity = verbosity

        # Check dimensionality
        if len(self.sensdata) != len(self.pointdata):
            raise ValueError("The sensitivity data does not match the point data.")

        super().__init__(**kwargs)

    def _transcribe_cells(self, parameters: List[str]) -> List[Cell]:
        """Transcribes the cells from Components.i.plt files into
        PySAGAS Cell objects.

        Parameters
        -----------
        parameters : List[str]
            A list of the geometric design parameters.

        Returns
        --------
        cells : List[Cell]
            A list of all transcribed cells.

        See Also
        --------
        pysagas.geometry.Cell
        """
        coordinates = self.pointdata[["Points_0", "Points_1", "Points_2"]]
        cell_vertex_ids = self.celldata[
            ["Point Index 0", "Point Index 1", "Point Index 2"]
        ]

        # Construct cells
        if self.verbosity > 0:
            print("\nTranscribing cells from Components.i.plt...")
            pbar = tqdm(
                total=len(self.celldata.index),
                position=0,
                leave=True,
            )
        cells = []
        degen_cells = 0
        for cell in self.celldata.index:
            vertex_ids = cell_vertex_ids.loc[cell].values
            cell_v_coords = [coordinates.loc[v_id].values for v_id in vertex_ids]
            vertices = [Vector.from_coordinates(c) for c in cell_v_coords]

            # Extract sensitivity information for the cell
            vertex_sensitivity_info = [self.sensdata.loc[v_id] for v_id in vertex_ids]

            dvdp = np.empty((9, len(parameters)))
            for i, p in enumerate(parameters):
                r = 0
                for v in vertex_sensitivity_info:
                    for c in ["x", "y", "z"]:
                        dvdp[r, i] = v[f"d{c}d{p}"]
                        r += 1

            # Create Cell
            try:
                newcell = Cell.from_points(vertices)
            except DegenerateCell:
                # Bad cell, skip it
                degen_cells += 1
                if self.verbosity > 0:
                    pbar.update(1)

                    if self.verbosity > 2:
                        print("\033[1mWarning\033[0m: Degenerate cell.")
                continue

            # Add geometry sensitivity information
            newcell._add_sensitivities(np.array(dvdp))

            # Add flow state at cell
            cell_velocity = Vector.from_coordinates(
                self.celldata.loc[cell][["U", "V", "W"]].values
            )
            cell_flowstate = FlowState(
                mach=self.celldata.loc[cell]["M"],
                pressure=self.celldata.loc[cell]["p"],
                temperature=self.celldata.loc[cell]["T"],
                direction=cell_velocity,
            )
            newcell.flowstate = cell_flowstate

            # Append new Cell
            cells.append(newcell)

            # Update progress bar
            if self.verbosity > 0:
                pbar.update(1)

        if self.verbosity > 0:
            pbar.close()
            print("Done.")

            if self.verbosity > 1:
                if degen_cells > 0:
                    print(f"{100*degen_cells/len(self.celldata):.2f}% degenerate cells")

        return cells



import numpy as np

np.seterr(all="ignore")
import gdtk.ideal_gas_flow as igf
from pysagas.geometry import Vector, Cell
from pysagas.flow import GasState, FlowState
from pysagas.sensitivity.calculator import SensitivityCalculator


def calculate_pressures(flow: FlowState, theta: float) -> float:
    """Calculates the pressure from a flow state and deflecion angle
    using ideal gas oblique shock theory.

    Parameters
    ----------
    flow : FlowState
        The flow state.

    theta : float
        The deflection angle (radians).

    Returns
    --------
    P2 : float
        The pressure behind the oblique shock.
    """
    beta = igf.beta_obl(M1=flow.M, theta=abs(theta), g=flow.gamma, tol=1.0e-6)
    P2_P1 = igf.p2_p1_obl(flow.M, beta, g=flow.gamma)
    P2 = P2_P1 * flow.P
    return P2


def run_main():
    # Define freestream flow state
    freestream = GasState(mach=6, pressure=700, temperature=70)

    # Define geometric parameters
    theta = np.radians(10)
    length = 0.01
    width = 0.02
    parameters = [theta, length, width]

    beta = igf.beta_obl(freestream.M, abs(theta), g=freestream.gamma, tol=1.0e-6)
    M_ramp = igf.M2_obl(freestream.M, beta, theta, g=freestream.gamma)
    T2_T1 = igf.T2_T1_obl(freestream.M, beta, g=freestream.gamma)

    P_ramp = calculate_pressures(flow=freestream, theta=theta)

    T_ramp = T2_T1 * freestream.T
    rho_ramp = P_ramp / (freestream.R * T_ramp)
    a_ramp = np.sqrt(freestream.gamma * freestream.R * T_ramp)
    vel_ramp = M_ramp * a_ramp

    # Step 1: Work out geometry position sensitivities
    #     p00-------p10
    #      |  A   /  |
    #      |    /    |        yo-->x
    #      |  /   B  |         |
    #     p01-------p11       \/z

    p00 = Vector(0.0, 0.0, 0.0)
    p01 = Vector(0.0, 0.0, parameters[2])
    p10 = Vector(
        parameters[1] * np.cos(parameters[0]),
        parameters[1] * np.sin(parameters[0]),
        0.0,
    )
    p11 = Vector(
        parameters[1] * np.cos(parameters[0]),
        parameters[1] * np.sin(parameters[0]),
        parameters[2],
    )

    # Calculate cell point-to-parameter sensitivities
    p00_sense = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
    p01_sense = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 1.0]])
    p10_sense = np.array(
        [
            [parameters[1] * -np.sin(parameters[0]), np.cos(parameters[0]), 0.0],
            [parameters[1] * np.cos(parameters[0]), np.sin(parameters[0]), 0.0],
            [0.0, 0.0, 0.0],
        ]
    )
    p11_sense = np.array(
        [
            [parameters[1] * -np.sin(parameters[0]), np.cos(parameters[0]), 0.0],
            [parameters[1] * np.cos(parameters[0]), np.sin(parameters[0]), 0.0],
            [0.0, 0.0, 1.0],
        ]
    )

    # Cell vertex sensitivity to params (dv/dp)
    pos_sense_A = np.vstack((p00_sense, p01_sense, p10_sense))
    pos_sense_B = np.vstack((p01_sense, p11_sense, p10_sense))

    # Define cells
    A = Cell(p00, p01, p10)
    B = Cell(p01, p11, p10)

    # Add sensitivities
    A._add_sensitivities(pos_sense_A)
    B._add_sensitivities(pos_sense_B)

    # Calculate velocity vector
    vel_ramp_vector = [vel_ramp * np.cos(theta), vel_ramp * np.sin(theta), 0.0]

    # Define ramp flow state
    ramp_velocity_vector = Vector.from_coordinates(vel_ramp_vector)
    ramp_fs = FlowState(
        mach=M_ramp, pressure=P_ramp, temperature=T_ramp, direction=ramp_velocity_vector
    )

    # Attribute the ramp flow state to each Cell
    A.flowstate = ramp_fs
    B.flowstate = ramp_fs

    # Calculate force sensitivity
    F_sense, _ = SensitivityCalculator.net_sensitivity(cells=[A, B])

    # Calculate error with finite differencing method
    fd_F_sense = calc_fd_sens(parameters, freestream)
    errors = 100 * np.nan_to_num(
        np.divide((fd_F_sense.T - F_sense), fd_F_sense.T), posinf=0, neginf=0
    )

    # Test against expected values (error < 10%)
    assert np.max(np.abs(errors)) < 10, "Adjoints inaccurately calculated"


def calculate_force_vector(P: float, n: np.array, A: float) -> np.array:
    """Calculates the force vector components, acting on a
    surface defined by its area and normal vector.

    Parameters
    ----------
    P : float
        The pressure (Pa).

    n : np.array
        The normal vector.

    A : float
        The reference area (m^2).

    Returns
    --------
    forces : np.array
        The force components.
    """
    F_x = A * P * np.dot(n, np.array([-1, 0, 0]))
    F_y = A * P * np.dot(n, np.array([0, -1, 0]))
    F_z = A * P * np.dot(n, np.array([0, 0, -1]))

    return [F_x, F_y, F_z]


def calc_fd_sens(parameters: list, freestream: GasState):
    # Perturbation parameters
    dp = 0.001  # 0.1% perturbation

    # Calculate sensitivities using finite difference
    pressure_sensitivities_finitediff = np.zeros((1, 3))
    force_sensitivities_finitediff = np.zeros((3, 3))
    for i, _ in enumerate(parameters):
        p_high = parameters.copy()
        p_low = parameters.copy()
        delta_p = dp * abs(parameters[i])
        p_high[i] = p_high[i] + delta_p
        p_low[i] = p_low[i] - delta_p

        pressure_high = calculate_pressures(freestream, p_high[0])
        pressure_low = calculate_pressures(freestream, p_low[0])

        # Pressure sensitivities
        pressure_sensitivities_finitediff[:, i] = (pressure_high - pressure_low) / (
            2 * delta_p
        )

        # Force sensitivities
        n_ramp_high = np.array([-1 * np.sin(p_high[0]), np.cos(p_high[0]), 0])
        n_ramp_low = np.array([-1 * np.sin(p_low[0]), np.cos(p_low[0]), 0])
        area_high = p_high[1] * p_high[2]
        area_low = p_low[1] * p_low[2]

        forces_high = calculate_force_vector(
            P=pressure_high, n=n_ramp_high, A=area_high
        )
        forces_low = calculate_force_vector(P=pressure_low, n=n_ramp_low, A=area_low)

        force_sensitivities_finitediff[:, i] = [
            (fh - fl) / (2 * delta_p) for fh, fl in zip(forces_high, forces_low)
        ]

    return force_sensitivities_finitediff


def test_ramp():
    run_main()


if __name__ == "__main__":
    run_main()



import numpy as np
from .geometry.vector import Vector


class GasState:
    """An ideal gas state defined by Mach number, pressure and
    temperature.
    """

    gamma = 1.4
    R = 287  # J/kg·K.3

    def __init__(
        self, mach: float, pressure: float, temperature: float, gamma: float = 1.4
    ) -> None:
        """Define a new gas state.

        Parameters
        -----------
        mach : float
            The flow Mach number.

        pressure : float
            The flow pressure (Pa).

        temperature : float
            The flow temperature (K).

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.
        """
        # Assign properties
        self._T = temperature
        self._P = pressure
        self._M = mach
        self._gamma = gamma

    def __str__(self) -> str:
        return f"Mach {self.M} flow condition with P = {self.P}, T = {self.T}."

    def __repr__(self) -> str:
        return f"Flow(M={self.M}, P={self.P}, T={self.T})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, GasState):
            raise Exception(f"Cannot compare {type(other)} to GasState.")
        return (
            (self._T == other._T)
            & (self._P == other._P)
            & (self._M == other._M)
            & (self._gamma == other._gamma)
        )

    @property
    def T(self):
        return self._T

    @property
    def P(self):
        return self._P

    @property
    def M(self):
        return self._M

    @property
    def a(self):
        return (self.gamma * self.R * self.T) ** 0.5

    @property
    def rho(self):
        return self.P / (self.R * self.T)

    @property
    def v(self):
        return self.M * self.a

    @property
    def q(self):
        return 0.5 * self.rho * self.v**2

    @property
    def gamma(self):
        return self._gamma


class FlowState(GasState):
    """An ideal gas state defined by Mach number, pressure and
    temperature, with a flow direction.
    """

    def __init__(
        self,
        mach: float,
        pressure: float,
        temperature: float,
        direction: Vector = None,
        aoa: float = 0.0,
        gamma: float = 1.4,
    ) -> None:
        """Define a new flow state.

        Parameters
        -----------
        mach : float
            The flow Mach number.

        pressure : float
            The flow pressure (Pa).

        temperature : float
            The flow temperature (K).

        direction : Vector, optional
            The direction vector of the flow. The default is Vector(1,0,0).

        aoa : float, optional
            The angle of attack of the flow. The default is 0.0 (specified in
            degrees).

        gamma : float, optional
            The ratio of specific heats. The default is 1.4.
        """
        super().__init__(mach, pressure, temperature, gamma)
        if direction:
            # Use direction provided
            self.direction = direction.unit
        else:
            # Use AoA to calculate direction
            self.direction = Vector(1, 1 * np.tan(np.deg2rad(aoa)), 0).unit

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FlowState):
            raise Exception(f"Cannot compare {type(other)} to FlowState.")
        same_gs = super().__eq__(other)
        return same_gs & (self.direction == other.direction)

    @property
    def vx(self):
        return self.Vector.x

    @property
    def vy(self):
        return self.Vector.y

    @property
    def vz(self):
        return self.Vector.z

    @property
    def vec(self):
        return self.Vector.vec

    @property
    def Vector(self) -> Vector:
        return self.direction * self.v

    @property
    def aoa(self):
        aoa = np.rad2deg(np.arctan(self.vec[1] / self.vec[0]))
        return round(aoa, 6)


if __name__ == "__main__":
    flow = FlowState(6, 700, 70)



import numpy as np
import pandas as pd
from tqdm import tqdm
from typing import List, Optional
from pysagas.geometry import Cell


def add_sens_data(
    cells: List[Cell],
    data: pd.DataFrame,
    verbosity: Optional[int] = 1,
    match_tolerance: Optional[float] = 1e-5,
    rounding_tolerance: Optional[float] = 1e-8,
    force: Optional[bool] = False,
) -> float:
    """Appends shape sensitivity data to a list of Cell objects.

    Parameters
    ----------
    cells : List[Cell]
        A list of cell objects.

    data : pd.DataFrame
        The sensitivity data to be transcribed onto the cells.

    verbosity : int, optional
        The verbosity. The default is 1.

    match_tolerance : float, optional
        The precision tolerance for matching point coordinates. The
        default is 1e-5.

    rounding_tolerance : float, optional
        The tolerance to round data off to. The default is 1e-8.

    force : bool, optional
        Force the sensitivity data to be added, even if a cell
        already has sensitivity data. This can be used if new
        data is being used, or if the append is being repeated with
        a different matching tolerance. The default is False.

    Returns
    --------
    match_fraction : float
        The fraction of points matched.
    """
    # Extract parameters
    parameters = []
    param_cols = data.columns[3:]
    for i in range(int(len(param_cols) / 3)):
        parameters.append(param_cols[int(i * 3)].split("dxd")[-1])

    # Construct progress bar
    if verbosity > 0:
        print()
        desc = "Adding geometry-parameter sensitivity data to cells"
        pbar = tqdm(
            total=len(cells),
            desc=desc,
            position=0,
            leave=True,
        )

    matched_points = 0
    total_points = 0
    skipped_cells = 0
    total_cells = 0
    for cell in cells:
        # Check if cell already has sensitivity data
        total_cells += 1
        if cell.dvdp is None or force:
            # Initialise sensitivity
            dvdp = np.zeros((9, len(parameters)))
            for i, point in enumerate([cell.p0, cell.p1, cell.p2]):
                match_x = (point.x - data["x"]).abs() < match_tolerance
                match_y = (point.y - data["y"]).abs() < match_tolerance
                match_z = (point.z - data["z"]).abs() < match_tolerance

                match = match_x & match_y & match_z
                try:
                    # Get match
                    matched_data = data[match].iloc[0][param_cols]

                    # Round off infinitesimally small values
                    matched_data[abs(matched_data) < rounding_tolerance] = 0

                    for j, p in enumerate(parameters):
                        # For each parameter (column)
                        for k, c in enumerate(["x", "y", "z"]):
                            dvdp[3 * i + k, j] = matched_data[f"d{c}d{p}"]

                    # Update match count
                    matched_points += 1

                except IndexError:
                    # No match found, leave as zero sensitivity
                    pass

                # Update total_points
                total_points += 1

            cell._add_sensitivities(np.array(dvdp))

        else:
            # Skip this cell
            skipped_cells += 1

        # Update progress bar
        if verbosity > 0:
            pbar.update(1)

    try:
        match_fraction = matched_points / total_points
    except:
        match_fraction = 1

    if verbosity > 0:
        pbar.close()
        print(
            f"Done - matched {100*match_fraction:.2f}% of points "
            + f"(skipped {100*skipped_cells/total_cells:.2f}% of cells)."
        )

    # TODO - allow dumping data to file

    return match_fraction



from __future__ import annotations
import os
import copy
import meshio
import numpy as np
import pandas as pd
from pysagas import banner
from abc import ABC, abstractmethod
from typing import List, Optional, Dict
from pysagas import Cell, FlowState, Vector


class AbstractFlowSolver(ABC):
    """Abstract interface for flow solvers."""

    method = None

    @abstractmethod
    def __init__(self, **kwargs) -> None:
        pass

    def __repr__(self) -> str:
        return f"PySAGAS {self.method} flow solver"

    def __str__(self) -> str:
        return f"PySAGAS {self.method} flow solver"

    @property
    @abstractmethod
    def method(self):
        # This is a placeholder for a class variable defining the solver method
        pass

    @abstractmethod
    def solve(self, *args, **kwargs):
        pass


class FlowSolver(AbstractFlowSolver):
    """Base class for CFD flow solver."""

    def __init__(
        self,
        cells: List[Cell],
        freestream: Optional[FlowState] = None,
        verbosity: Optional[int] = 1,
    ) -> None:
        """Instantiates the flow solver.

        Parameters
        ----------
        cells : list[Cell]
            A list of the cells describing the geometry.

        freestream : Flowstate, optional
            The free-stream flow state. The default is the freestream provided
            upon instantiation of the solver.

        verbosity : int, optional
            The verbosity of the solver. The default is 1.
        """

        # TODO - allow providing a geometry parser, eg. tri file parser for Cart3D,
        # or a (yet to be implemented) STL parser.

        self.cells = cells
        self.freestream = freestream
        self.verbosity = verbosity
        self._last_solve_freestream: FlowState = None
        self._last_sens_freestream: FlowState = None

        # Results
        self.flow_result: FlowResults = None
        self.flow_sensitivity: SensitivityResults = None

        # Print banner
        if self.verbosity > 0:
            banner()
            print(f"\033[4m{self.__repr__()}\033[0m".center(50, " "))

    def solve(
        self,
        freestream: Optional[FlowState] = None,
        mach: Optional[float] = None,
        aoa: Optional[float] = None,
    ) -> bool:
        """Run the flow solver.

        Parameters
        ----------
        freestream : Flowstate, optional
            The free-stream flow state. The default is the freestream provided
            upon instantiation of the solver.

        mach : float, optional
            The free-stream Mach number. The default is that specified in
            the freestream flow state.

        aoa : float, optional
            The free-stream angle of attack. The default is that specified in
            the freestream flow state.

        Returns
        -------
        result : FlowResults
            The flow results.

        Raises
        ------
        Exception : when no freestream can be found.
        """

        if not freestream:
            # No freestream conditions specified
            if not self.freestream:
                # No conditions provided on instantiation either
                raise Exception("Please specify freestream conditions.")
            else:
                # Use nominal freestream as base
                freestream = copy.copy(self.freestream)
                if mach:
                    # Update mach number
                    freestream._M = mach

                if aoa:
                    # Update flow direction
                    flow_direction = Vector(1, 1 * np.tan(np.deg2rad(aoa)), 0).unit
                    freestream.direction = flow_direction

        else:
            # Solve-specific freestream conditions provided
            if mach or aoa:
                print("Using freestream provided; ignoring Mach/aoa provided.")

        # Check if already solved
        if self._last_solve_freestream and freestream == self._last_solve_freestream:
            return True
        else:
            # Update last solve freestream and continue
            self._last_solve_freestream = freestream
            return False

    def solve_sens(
        self,
        freestream: Optional[FlowState] = None,
        mach: Optional[float] = None,
        aoa: Optional[float] = None,
    ) -> SensitivityResults:
        """Run the flow solver to obtain sensitivity information.

        Parameters
        ----------
        freestream : Flowstate, optional
            The free-stream flow state. The default is the freestream provided
            upon instantiation of the solver.

        Mach : float, optional
            The free-stream Mach number. The default is that specified in
            the freestream flow state.

        aoa : float, optional
            The free-stream angle of attack. The default is that specified in
            the freestream flow state.

        Returns
        -------
        SensitivityResults

        Raises
        ------
        Exception : when no freestream can be found.
        """
        if not freestream:
            # No freestream conditions specified
            if not self.freestream:
                # No conditions provided on instantiation either
                raise Exception("Please specify freestream conditions.")
            else:
                # Use nominal freestream as base
                freestream = copy.copy(self.freestream)
                if mach:
                    # Update mach number
                    freestream._M = mach

                if aoa:
                    # Update flow direction
                    flow_direction = Vector(1, 1 * np.tan(np.deg2rad(aoa)), 0).unit
                    freestream.direction = flow_direction

        else:
            # Solve-specific freestream conditions provided
            if mach or aoa:
                print("Using freestream provided; ignoring Mach/aoa provided.")

        # Check if already solved
        if self._last_sens_freestream and freestream == self._last_sens_freestream:
            return True
        else:
            # Update last solve freestream and continue
            self._last_sens_freestream = freestream
            return False

    def save(self, name: str, attributes: Dict[str, list]):
        """Save the solution to VTK file format. Note that the
        geometry mesh must have been loaded from a parser which
        includes connectivity information (eg. PyMesh or TRI).

        Parameters
        ----------
        name : str
            The filename prefix of the desired output file.

        attributes : Dict[str, list]
            The attributes dictionary used to initialise the file writer.
            Note that this is not required when calling `solve` from a
            child FlowSolver.
        """
        # Import PyMesh
        try:
            import pymesh
        except ModuleNotFoundError:
            raise Exception(
                "Could not find pymesh. Please follow the "
                + "installation instructions at "
                + "https://pymesh.readthedocs.io/en/latest/installation.html"
            )

        # Construct vertices and faces
        faces = []
        vertex_faces = {}
        for cell in self.cells:
            # Get faces
            faces.append(cell._face_ids)

            # Get attributes
            for a in attributes:
                attributes[a].append(cell.attributes.get(a))

            # Get vertices
            for i, vid in enumerate(cell._face_ids):
                vertex_faces[vid] = cell.vertices[i]

        # Sort vertices
        sorted_vertices = dict(sorted(vertex_faces.items()))

        # Note: vertex attributes can be treated the same way as
        # the vertices - they must be sorted.

        # Finally
        vertices = np.array([v for v in sorted_vertices.values()])
        faces = np.array(faces)

        # Convert back to mesh object
        mesh_obj = pymesh.form_mesh(vertices, faces)

        # Try add attributes to the cells
        for a, v in attributes.items():
            mesh_obj.add_attribute(a)
            mesh_obj.set_attribute(a, np.array(v))

        # Save mesh
        pymesh.save_mesh(
            f"{name}.ply", mesh_obj, *mesh_obj.get_attribute_names(), ascii=True
        )

        # Convert to VTK format
        m = meshio.read(f"{name}.ply")
        m.write(f"{name}.vtk")

        # Remove .ply file
        os.remove(f"{name}.ply")

    @staticmethod
    def body_to_wind(v: Vector, aoa: float):
        """Converts a vector from body axes to wind axes.

        Parameters
        ----------
        v : Vector
            The result to transform.

        aoa : float
            The angle of attack, specified in degrees.

        Returns
        --------
        r : Vector
            The transformed result.
        """
        # TODO - use euler matrices
        A = v.x  # axial
        N = v.y  # normal

        L = N * np.cos(np.deg2rad(aoa)) - A * np.sin(np.deg2rad(aoa))
        D = N * np.sin(np.deg2rad(aoa)) + A * np.cos(np.deg2rad(aoa))

        # Construct new result
        r = Vector(x=D, y=L, z=v.z)

        return r

    def sweep(self, aoa_range: list[float], mach_range: list[float]) -> pd.DataFrame:
        """Evaluate the aerodynamic coefficients over a sweep of angle
        of attack and mach numbers.

        Parameters
        -----------
        aoa_range : list[float]
            The angles of attack to evaluate at, specified in degrees.

        mach_range : list[float]
            The Mach numbers to evaluate at.

        Returns
        --------
        results : pd.DataFrame
            A Pandas DataFrame of the aerodynamic coefficients at each angle of
            attack and Mach number combination.
        """
        # TODO - add bool to evaluate gradients on the sweep
        # Run sweep
        i = 0
        results = pd.DataFrame()
        for aoa in aoa_range:
            for mach in mach_range:
                flowresult: FlowResults = self.solve(aoa=aoa, Mach=mach)
                CL, CD, Cm = flowresult.coefficients()

                coefficients_df = pd.DataFrame(
                    data={"aoa": aoa, "mach": mach, "CL": CL, "CD": CD, "Cm": Cm},
                    index=[i],
                )

                # Append results
                results = pd.concat(
                    [
                        results,
                        coefficients_df,
                    ]
                )

                # Iterate index
                i += 1

        return results


class FlowResults:
    """A class containing the aerodynamic force and moment
    information with respect to design parameters.

    Attributes
    ----------
    freestream : FlowState
        The freestream flow state.

    net_force : pd.DataFrame
        The net force in cartesian coordinate frame (x,y,z).

    m_sense : pd.DataFrame
        The net moment in cartesian coordinate frame (x,y,z).
    """

    def __init__(
        self, freestream: FlowState, net_force: Vector, net_moment: Vector
    ) -> None:
        self.freestream = freestream
        self.net_force = net_force
        self.net_moment = net_moment

        # Calculate angle of attack
        self.aoa = np.rad2deg(
            np.arctan(freestream.direction.y / freestream.direction.x)
        )

    def __str__(self) -> str:
        return f"Net force = {self.net_force} N"

    def __repr__(self) -> str:
        return self.__str__()

    def coefficients(self, A_ref: Optional[float] = 1.0, c_ref: Optional[float] = 1.0):
        """Calculate the aerodynamic coefficients CL, CD and Cm.

        Parameters
        -----------
        A_ref : float, optional
            The reference area (m^2). The default is 1 m^2.

        c_ref : float, optional
            The reference length (m). The default is 1 m.

        Returns
        -------
        C_L, C_D, C_m
        """
        w = FlowSolver.body_to_wind(v=self.net_force, aoa=self.aoa)
        C_L = w.y / (self.freestream.q * A_ref)
        C_D = w.x / (self.freestream.q * A_ref)

        mw = FlowSolver.body_to_wind(v=self.net_moment, aoa=self.aoa)
        C_m = -mw.z / (self.freestream.q * A_ref * c_ref)
        return C_L, C_D, C_m

    @property
    def coef(self):
        CL, CD, Cm = self.coefficients()
        print(f"CL = {CL:.3f}\nCD = {CD:.3f}\nCm = {Cm:.3f}")


class SensitivityResults:
    """A class containing the aerodynamic force and moment sensitivity
    information with respect to design parameters.

    Attributes
    ----------
    freestream : FlowState
        The freestream flow state.

    f_sense : pd.DataFrame
        The force sensitivities in cartesian coordinate frame (x,y,z).

    m_sense : pd.DataFrame
        The moment sensitivities in cartesian coordinate frame (x,y,z).
    """

    def __init__(
        self,
        f_sens: pd.DataFrame,
        m_sens: pd.DataFrame,
        freestream: Optional[FlowState] = None,
    ) -> None:
        self.f_sens = f_sens
        self.m_sens = m_sens
        self.freestream = freestream

    def __str__(self) -> str:
        s1 = f"Force sensitivties:\n{self.f_sens}"
        s2 = f"Moment sensitivties:\n{self.m_sens}"

        return f"{s1}\n\n{s2}"

    def __repr__(self) -> str:
        return self.__str__()

    def coefficients(
        self, A_ref: Optional[float] = 1.0, c_ref: Optional[float] = 1.0
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Calculate the aerodynamic coefficients CL, CD and Cm.

        Parameters
        -----------
        A_ref : float, optional
            The reference area (m^2). The default is 1 m^2.

        c_ref : float, optional
            The reference length (m). The default is 1 m.

        Returns
        -------
        A tuple (cf_sens, cm_sens) containing the force and moment coefficient sensitivities.
        """
        # Non-dimensionalise
        cf_sens: pd.DataFrame = self.f_sens / (self.freestream.q * A_ref)
        cm_sens: pd.DataFrame = self.m_sens / (self.freestream.q * A_ref * c_ref)

        # Translate to aero frame
        # TODO - review if this might swap value order on overwrite of columns
        aoa = self.freestream.aoa
        cf_sens["dFy/dp"] = cf_sens["dFy/dp"] * np.cos(np.deg2rad(aoa)) - cf_sens[
            "dFx/dp"
        ] * np.sin(np.deg2rad(aoa))
        cf_sens["dFx/dp"] = cf_sens["dFy/dp"] * np.sin(np.deg2rad(aoa)) + cf_sens[
            "dFx/dp"
        ] * np.cos(np.deg2rad(aoa))

        # Rename columns
        cf_sens.rename(
            columns={"dFx/dp": "dCD", "dFy/dp": "dCL", "dFz/dp": "dCZ"}, inplace=True
        )
        # cm_sens.rename(columns={"dMx/dp": "dCl", "dMy/dp": "dCn", "dMz/dp": "dCm"}, inplace=True)

        return cf_sens, cm_sens



import os
import numpy as np
from pysagas import banner
from numpy.typing import ArrayLike
from typing import Union, Optional, Dict, Any
from pyoptsparse import Optimizer, Optimization


class ShapeOpt:
    """PySAGAS Shape Optimisation wrapper for pyOptSparse."""

    def __init__(
        self,
        optimiser: Optimizer,
        working_dir: str,
        optimiser_options: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Initialise PySAGAS Shape Optimiser.

        Parameters
        ----------
        optimiser : Optimizer
            The pyoptsparse Optimizer object of choice.

        working_dir : str
            The name of the working directory.

        optimiser_options : dict, optional
            The options to pass to the optimiser.
        """
        self.opt_problem: Optimization
        self.optimiser = optimiser(options=optimiser_options)

        # Prepare working directory
        if not os.path.exists(working_dir):
            os.mkdir(working_dir)

    def add_variables(
        self,
        parameters_dict: Optional[dict] = None,
        name: Optional[str] = None,
        n: Optional[int] = None,
        vartype: Optional[str] = "c",
        initial: Optional[ArrayLike] = None,
        lower_bounds: Optional[Union[ArrayLike, Dict]] = None,
        upper_bounds: Optional[Union[ArrayLike, Dict]] = None,
        **kwargs,
    ):
        """Adds variables to the optimisation problem.

        Parameters
        ----------
        parameters_dict : dict, optional
            The dictionary of nominal parameter values. This can be provided
            instead of adding each parameter individually with the 'name'
            argument.

        name : str
            Name of variable group. This name should be unique across all the design variable groups

        n : int, optional
            Number of design variables in this variable group.

        varType : str.
            String representing the type of variable. Suitable values for type
            are: 'c' for continuous variables, 'i' for integer values and
            'd' for discrete selection.

        value : scalar or array.
            Starting value for design variables. If it is a a scalar, the same
            value is applied to all 'nVars' variables. Otherwise, it must be
            iterable object with length equal to 'nVars'.

        lower : scalar or array.
            Lower bound of variables. Scalar/array usage is the same as value
            keyword

        upper : scalar or array.
            Upper bound of variables. Scalar/array usage is the same as value
            keyword
        """

        def check_bounds(
            parameters: Dict[str, float], bounds: Dict[str, float], sign: int
        ):
            for p, v in parameters.items():
                # Get bound for this parameter
                b = bounds.get(p)
                if np.sign(v - b) != np.sign(sign) and np.sign(v - b) != 0:
                    if np.sign(sign) < 0:
                        bound = ">"
                    else:
                        bound = "<"
                    raise Exception(
                        f"Invalid bounds on {p}: " + f"{v:.5f} {bound} {b:.5f}."
                    )

        if parameters_dict:
            # Check bounds
            if lower_bounds is None:
                lower_bounds = {}
            else:
                # Make sure bounds are valid
                check_bounds(parameters_dict, lower_bounds, 1)

            if upper_bounds is None:
                upper_bounds = {}
            else:
                # Make sure bounds are valid
                check_bounds(parameters_dict, upper_bounds, -1)

            # Unpack parameters dict
            for param, value in parameters_dict.items():
                self.opt_problem.addVarGroup(
                    name=param,
                    nVars=1,
                    varType=vartype,
                    value=value,
                    lower=lower_bounds.get(param),
                    upper=upper_bounds.get(param),
                    **kwargs,
                )

        else:
            self.opt_problem.addVarGroup(
                name=name,
                nVars=n,
                varType=vartype,
                value=initial,
                lower=lower_bounds,
                upper=upper_bounds,
                **kwargs,
            )

    def add_constriants(
        self,
        name: str,
        n: Optional[int] = 1,
        lower: Optional[Union[ArrayLike, Dict]] = None,
        upper: Optional[Union[ArrayLike, Dict]] = None,
        scale: Optional[float] = 1,
    ):
        """Adds constraints to the optimisation problem.

        Parameters
        -----------
        name : str
            The name key of the constraint being added. This must appear
            in the dictionary returned by the objective and Jacobian callback
            returns.

        nCon : int
            The number of constraints in this group

        lower : scalar or array
            The lower bound(s) for the constraint. If it is a scalar,
            it is applied to all nCon constraints. If it is an array,
            the array must be the same length as nCon.

        upper : scalar or array
            The upper bound(s) for the constraint. If it is a scalar,
            it is applied to all nCon constraints. If it is an array,
            the array must be the same length as nCon.

        scale : scalar or array
            A scaling factor for the constraint. It is generally
            advisable to have most optimization constraint around the
            same order of magnitude.
        """
        self.opt_problem.addConGroup(
            name=name,
            nCon=n,
            lower=lower,
            upper=upper,
            scale=scale,
            linear=None,
            wrt=None,
            jac=None,
        )

    def add_objective(self, name: str):
        """Adds an objective to the optimisation problem.

        Parameters
        -----------
        name : str
            The name key of the objective being added. This must appear
            in the dictionary returned by the objective and Jacobian callback
            returns.
        """
        self.opt_problem.addObj(name)

    def run(self, hotstart_file: str = None):
        """Run ShapeOpt.

        Parameters
        -----------
        hotstart_file : str, optional
            The filepath to the history file, used to hot start the
            optimiser.
        """
        # Print banner
        banner()
        print("\033[4mPySAGAS Shape Optimisation\033[0m".center(50, " "))

        # Run optimiser
        self.sol = self.optimiser(
            self.opt_problem,
            storeHistory="history.hst",
            hotStart=hotstart_file,
        )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

sys.path.insert(0, os.path.abspath("./../.."))


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "PySAGAS"
copyright = "2022, Kieran Mackle, Ingo Jahn"
author = "Kieran Mackle, Ingo Jahn"
release = "0.1.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "myst_parser",
    "sphinx_inline_tabs",
]

templates_path = ["_templates"]
exclude_patterns = ["build"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_favicon = "favicon.ico"
html_show_sourcelink = True

myst_enable_extensions = ["dollarmath", "amsmath"]


# -- Options for autodoc ----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#configuration

# Automatically extract typehints when specified and place them in
# descriptions of the relevant function/method.
autodoc_typehints = "description"

# Don't show class signature with the class' name.
autodoc_class_signature = "separated"



import os
import time
import shutil
import subprocess
import numpy as np
from pysagas import FlowState, Vector
from typing import Dict, List, Optional
from pysagas.optimisation.cart3d.utilities import C3DPrep
from pysagas.sensitivity import Cart3DSensitivityCalculator
from pysagas.cfd.solver import FlowSolver, FlowResults, SensitivityResults


class Cart3D(FlowSolver):
    method = "Cart3D"

    _C3D_errors = [
        "==> ADAPT failed",
        "Check cart3d.out in AD_A_J for more clues",
        "==> adjointErrorEst_quad failed again, status = 1",
        "ERROR: CUBES failed",
        "ERROR: ADAPT failed with status = 1",
        "ERROR",
        "Try decreasing etol, increasing mesh growth or switching to 'a' cycles",
    ]
    _C3D_done = "Done aero.csh"

    def __init__(
        self,
        stl_files: List[str],
        aero_csh: str = "aero.csh",
        input_cntl: str = "input.cntl",
        freestream: Optional[FlowState] = None,
        verbosity: Optional[int] = 1,
        c3d_log_name: str = "C3D_log",
        c3d_info_file: str = None,
        write_config_xml: bool = True,
        ref_area: Optional[float] = 1.0,
        ref_length: Optional[float] = 1.0,
    ) -> None:
        """Instantiate the Cart3D solver wrapper.

        Parameters
        -----------
        stl_files : list[str]
            A list of filepaths to the STL files to intersect and
            simulate.

        aero_csh : str
            The filepath to the reference aero.csh file. The default
            option will look in the current working directory.

        c3d_log_name : str, optional
            The name to use for the Cart3D logfile. The default is C3D_log.

        write_config_xml : bool, optional
            A boolean flag to write the Cart3D Config.xml file when running
            comp2tri. If the geometry may be perturbed to a state where one
            component is no longer part of the wetted surface, writing the
            comp2tri file can cause set-up issues, and so it can be beneficial
            to turn it off. The default is True.
        """
        # Save STL files for processing
        self.stl_files = stl_files
        self._root_dir = os.getcwd()
        self._aerocsh_path = os.path.join(self._root_dir, aero_csh)
        self._inputcntl_path = os.path.join(self._root_dir, input_cntl)

        # Instantiate helper
        self.c3d_log_name = c3d_log_name
        self._prepper = C3DPrep(
            logfile=c3d_log_name,
            info_file=c3d_info_file,
            write_config=write_config_xml,
        )

        # Dimensional properties
        self.ref_area = ref_area
        self.ref_length = ref_length

        # Private attributes
        self._sim_dir = None
        self._killed = False
        self._donefile = None

        # Complete instantiation
        super().__init__(None, freestream, verbosity)

    def solve(
        self,
        freestream: Optional[FlowState] = None,
        mach: Optional[float] = None,
        aoa: Optional[float] = None,
    ) -> FlowResults:
        already_run = super().solve(freestream=freestream, mach=mach, aoa=aoa)
        if already_run:
            # Already have a result
            if self.verbosity > 1:
                print("Attention: this flowstate has already been solved.")
            result = self.flow_result

        else:
            # Get flow
            flow = self._last_solve_freestream
            loads = self._run_sim(mach=flow.M, aoa=flow.aoa)

            # TODO - need to be careful and aware of the reference coordinates here
            C_F = np.array(
                [loads["C_A-entire"], loads["C_N-entire"], loads["C_Y-entire"]]
            )
            C_M = np.array(
                [loads["C_M_x-entire"], loads["C_M_z-entire"], -loads["C_M_y-entire"]]
            )

            # Dimensionalise results
            net_force = C_F * flow.q * self.ref_area
            net_moment = C_M * flow.q * self.ref_area * self.ref_length

            # Convert to vectors
            net_force = Vector.from_coordinates(net_force)
            net_moment = Vector.from_coordinates(net_moment)

            # Construct results
            result = FlowResults(
                freestream=flow, net_force=net_force, net_moment=net_moment
            )

            # Save
            self.flow_result = result

            if self.verbosity > 1:
                print("Cart3D done.")

        return result

    @property
    def status(self):
        """Returns the status of the wrapper."""
        running, error = self._c3d_running(
            os.path.join(self._sim_dir, self.c3d_log_name), self._donefile
        )
        if error:
            print(f"Failed with error {error}.")
        else:
            if running:
                print("Running.")
            else:
                print("Complete.")

    def stop(self):
        """Stop running Cart3D."""
        self._killed = True
        if self.verbosity > 0:
            print("Kill signal received. Stopping Cart3D.")
        stopfile = os.path.join(self._sim_dir, "STOP")
        with open(stopfile, "w"):
            pass

    def solve_sens(
        self,
        sensitivity_filepath: str,
        freestream: Optional[FlowState] = None,
        mach: Optional[float] = None,
        aoa: Optional[float] = None,
    ) -> SensitivityResults:
        already_run = super().solve_sens(freestream=freestream, mach=mach, aoa=aoa)
        if already_run:
            # Already have a result
            if self.verbosity > 1:
                print("Attention: this flowstate sensitivity has already been solved.")
            result = self.flow_sensitivity

        else:
            # Get flow
            flow = self._last_sens_freestream

            # Check that nominal flow condition has been solved already
            if self._last_solve_freestream:
                # Solver has run before
                if self._last_solve_freestream != flow:
                    # Run flow solver first at the nominal flow conditions
                    if self.verbosity > 0:
                        print("Running flow solver to prepare sensitivities.")
                    self.solve(flow)
                else:
                    if self.verbosity > 0:
                        print("Using previously run flow solution.")
            else:
                # Solver has not run yet at all, run it now
                if self.verbosity > 0:
                    print("Running flow solver to prepare sensitivities.")
                self.solve(flow)

            # Create sensitivity wrapper
            components_filepath = os.path.join(
                self._sim_dir, "BEST/FLOW/Components.i.plt"
            )
            wrapper = Cart3DSensitivityCalculator(
                freestream=flow,
                sensitivity_filepath=sensitivity_filepath,
                components_filepath=components_filepath,
                verbosity=0,
            )

            # Calculate sensitivities
            df_f, df_m = wrapper.calculate()

            # Construct results
            result = SensitivityResults(
                f_sens=df_f,
                m_sens=df_m,
                freestream=flow,
            )

            # Save
            self.flow_sensitivity = result

        if self.verbosity > 0:
            print(result)

        return result

    def running(self) -> bool:
        """Returns True if Cart3D is actively running, else False."""
        if self._donefile:
            # Donefile has been assigned
            return self._c3d_running(
                os.path.join(self._sim_dir, self.c3d_log_name), self._donefile
            )[0]
        else:
            # Donefile has not been assigned
            return True

    def save(self, name: str, attributes: Dict[str, list]):
        raise NotImplementedError("Coming soon.")

    def _prepare_sim_dir(self, sim_dir_name: dir):
        """Prepares a new directory to run the simulation in."""
        # Make simulation directory
        sim_dir = os.path.join(self._root_dir, sim_dir_name)
        run_intersect = False
        components_filepath = os.path.join(sim_dir, "Components.i.tri")
        if not os.path.exists(sim_dir):
            os.mkdir(sim_dir)
            run_intersect = True
            intersected = False
        else:
            # Check for intersected file
            run_intersect = not os.path.exists(components_filepath)
            intersected = True

        # Copy STL files into sim directory
        for file in self.stl_files:
            shutil.copyfile(
                os.path.join(self._root_dir, file),
                os.path.join(sim_dir, file),
            )

        if self.verbosity > 0:
            print(f"Running simulation in {sim_dir}.")

        return run_intersect, intersected

    def _run_sim(self, mach: float, aoa: float):
        """Run the simulation."""
        # Prepare sim directory
        sim_dir = os.path.join(self._root_dir, f"M{mach}A{aoa}")
        self._sim_dir = sim_dir
        run_intersect, intersected = self._prepare_sim_dir(sim_dir)

        # Move into sim directory
        os.chdir(sim_dir)

        # Attempt simulation
        sim_success = False
        no_attempts = 3
        for attempt in range(no_attempts):
            # Run intersect
            if run_intersect:
                self._prepper._log(f"SHAPEOPT INTERSECT ATTEMPT {attempt+1}")
                intersected = self._prepper.intersect_stls(stl_files=self.stl_files)

            # Check for intersection
            if intersected:
                # Prepare rest of simulation directory
                if not os.path.exists(os.path.join(sim_dir, "input.cntl")):
                    # Move files to simulation directory (including Components.i.tri)
                    self._prepper.run_autoinputs()
                    # os.system(
                    #     f"mv *.tri Config.xml input.c3d preSpec.c3d.cntl {sim_dir} >> {self.c3d_log_name} 2>&1"
                    # )

                    # Copy sim files and permissions
                    for filename, fp in {
                        "input.cntl": self._inputcntl_path,
                        "aero.csh": self._aerocsh_path,
                    }.items():
                        shutil.copyfile(
                            os.path.join(self._root_dir, filename),
                            os.path.join(sim_dir, filename),
                        )
                        shutil.copymode(
                            os.path.join(self._root_dir, filename),
                            os.path.join(sim_dir, filename),
                        )

                    # Update flow conditions
                    new_input_cntl = os.path.join(sim_dir, "input.cntl")
                    self._edit_input_cntl(
                        mach=mach,
                        aoa=aoa,
                        original=self._inputcntl_path,
                        new=new_input_cntl,
                    )

                # Run Cart3D and await result
                target_adapt = self._infer_adapt(self._aerocsh_path)
                c3d_donefile = os.path.join(sim_dir, target_adapt, "FLOW", "DONE")
                self._donefile = c3d_donefile
                run_cmd = "./aero.csh restart"
                _restarts = 0
                if not os.path.exists(c3d_donefile):
                    with open(self.c3d_log_name, "a") as f:
                        subprocess.run(
                            run_cmd, shell=True, stdout=f, stderr=subprocess.STDOUT
                        )
                    while not os.path.exists(c3d_donefile):
                        # Wait...
                        time.sleep(5)

                        # Check for C3D failure
                        running, e = self._c3d_running(
                            c3d_log_name=self.c3d_log_name, donefile=c3d_donefile
                        )

                        # Check for kill status
                        if self._killed:
                            if self.verbosity > 1:
                                print("Kill signal received, exiting simulation loop.")
                            sim_success = False
                            break

                        if not running:
                            # C3D failed, try restart it
                            if _restarts > 3:
                                return False

                            if self.verbosity > 0:
                                print(f"Warning: Cart3D failed with error '{e}'")

                            f = open(self.c3d_log_name, "a")
                            subprocess.run(
                                run_cmd, shell=True, stdout=f, stderr=subprocess.STDOUT
                            )
                            f.close()
                            _restarts += 1

                sim_success = True
                break

        if sim_success:
            # Sim finished successfully, read loads file
            loads_dict = self._read_c3d_loads(
                os.path.join(sim_dir, "BEST/FLOW/loadsCC.dat"),
                tag="entire",
            )

        else:
            loads_dict = None

        # Change back to working directory
        os.chdir(self._root_dir)

        return loads_dict

    @staticmethod
    def _c3d_running(c3d_log_name: str, donefile: str) -> bool:
        """Watches the Cart3D log file to check for errors and return False
        if Cart3D has stopped running.

        Returns
        -------
        running : bool
            Whether Cart3D is still running.

        error : str
            The error message, if an error has occured.
        """
        if os.path.exists(donefile):
            return False, None

        # DONE file doesn't exist, read logfile for clues
        if os.path.exists(c3d_log_name):
            try:
                with open(c3d_log_name) as f:
                    # Get last line in log file
                    for line in f:
                        pass

                    # Check if if it is in the known errors
                    for e in Cart3D._C3D_errors:
                        if e in line:
                            return False, e

                    # Check if it says done
                    if Cart3D._C3D_done in line:
                        return False, None
            except:
                # Threading read error
                pass

        # No errors
        return True, None

    @staticmethod
    def _infer_adapt(aero_csh_fp: str) -> str:
        """Returns the target adapt number by reading the aero.csh file."""
        with open(aero_csh_fp, "r") as f:
            lines = f.readlines()

            for line in lines:
                if line.find("set n_adapt_cycles") != -1:
                    return f"adapt{int(line.split('=')[-1]):02d}"

    @staticmethod
    def _edit_input_cntl(mach: float, aoa: float, original: str, new: str):
        """Edits the Mach number and angle of attack in the input.cntl file."""
        with open(new, "w+") as new_file:
            with open(original, "r") as old_file:
                for line in old_file:
                    if line.startswith("Mach"):
                        line = f"Mach  {mach}\n"
                    elif line.startswith("alpha"):
                        line = f"alpha  {aoa}\n"

                    # Write line
                    new_file.write(line)

    @staticmethod
    def _read_c3d_loads(
        loadsCC_filepath: str,
        b_frame: bool = True,
        v_frame: bool = True,
        moments: bool = True,
        tag: str = None,
    ) -> dict:
        load_dict = {}
        with open(loadsCC_filepath, "r") as file:
            for line in file:
                if line[0] == "#":
                    # Commented line, skip
                    continue

                # Remove linebreaks and multiple spaces
                line = " ".join(line.split())
                words = line.split(":")

                if len(words) == 0 or len(words) == 1:  # skip if empty line
                    continue

                text = words[0]
                number = float(words[1])
                word = text.split(" ")
                _tag = word[0]

                if (not tag) or (tag is not None and tag == _tag):
                    coeff = word[-1]
                    short_coeff = coeff[1:4]
                    long_coeff = coeff[1:6]
                    if b_frame is True:
                        if short_coeff in ["C_A", "C_Y", "C_N"]:  # get force in b_frame
                            load_dict["{0}-{1}".format(short_coeff, tag)] = number
                    if v_frame is True:
                        if short_coeff in ["C_D", "C_S", "C_L"]:  # get force in v_frame
                            load_dict["{0}-{1}".format(short_coeff, tag)] = number
                    if moments is True:
                        if short_coeff in [
                            "C_l",
                            "C_m",
                            "C_n",
                            "C_M",
                        ]:  # get moment coeff
                            load_dict["{0}-{1}".format(short_coeff, tag)] = number
                        if b_frame is True:
                            if long_coeff in [
                                "C_M_x",
                                "C_M_y",
                                "C_M_z",
                            ]:  # get force in b_frame
                                load_dict["{0}-{1}".format(long_coeff, tag)] = number

        return load_dict



import os
import time
import glob
import pickle
import shutil
import subprocess
import numpy as np
import pandas as pd
import multiprocess as mp
from pysagas.flow import FlowState
from pysagas.cfd.cart3d import Cart3D
from hypervehicle.generator import Generator
from pyoptsparse import Optimizer, Optimization
from pysagas.optimisation.optimiser import ShapeOpt
from hypervehicle.utilities import SensitivityStudy
from pysagas.sensitivity.cart3d import Cart3DSensitivityCalculator
from pysagas.optimisation.cart3d.utilities import C3DPrep, combine_sense_data


np.seterr(all="ignore")


class Cart3DShapeOpt(ShapeOpt):
    """A wrapper to perform shape optimisation with Cart3D."""

    C3D_errors = [
        "==> ADAPT failed",
        "Check cart3d.out in AD_A_J for more clues",
        "==> adjointErrorEst_quad failed again, status = 1",
        "ERROR: CUBES failed",
        "ERROR: ADAPT failed with status = 1",
        "ERROR",
    ]

    def __init__(
        self,
        freestream: FlowState,
        optimiser: Optimizer,
        vehicle_generator: Generator,
        objective_callback: callable,
        jacobian_callback: callable,
        A_ref: float = 1.0,
        l_ref: float = 1.0,
        optimiser_options: dict = None,
        sensitivity_filename: str = "all_components_sensitivity.csv",
        working_dir_name: str = "working_dir",
        sim_directory_name: str = "simulation",
        basefiles_dir_name: str = "basefiles",
        c3d_log_name: str = "C3D_log",
        c3d_info_file: str = None,
        matching_tolerance: float = 1e-5,
        save_evolution: bool = True,
        write_config_xml: bool = True,
        sensitivity_kwargs: dict = None,
    ) -> None:
        """Initialise Cart3D Shape Optimiser.

        Parameters
        ----------
        a_inf : float
            The freestream speed of sound (m/s).

        rho_inf : float
            The freestream density (kg/m^3).

        V_inf : float
            The freestream velocity (m/s).

        A_ref : float
            The aerodynamic reference area (m^2).

        optimiser : Optimizer
            The pyoptsparse Optimizer object of choice.

        optimiser_options : dict, optional
            The options to pass to the optimiser. See the PyOptSparse
            documentation for solver-specific options. The default is None.

        vehicle_generator : Generator
            The vehicle generator object.

        objective_callback : callable
            The callback function to compute and return the objective function
            value and constraint violation values.

        jacobian_callback : callable
            The callback function to compute and return the Jacobian of the
            objective function value and constraint violation values.

        sensitivity_filename : str, optional
            The filename of the combined components sensitivities. The default
            is 'all_components_sensitivity.csv'.

        working_dir_name : str, optional
            The name of the working directory. The default is 'working_dir'.

        sim_dir_name : str, optional
            The name of the simulation directory. The default is 'simulation'.

        basefiles_dir_name : str, optional
            The name of the base files directory. The default is 'basefiles'.

        c3d_log_name : str, optional
            The name to use for the Cart3D logfile. The default is C3D_log.

        save_evolution : bool, optional
            A boolean flag to save geometry files of the evolving geometry.
            If True, the files will be saved to the evolution directory. The
            default is True.

        write_config_xml : bool, optional
            A boolean flag to write the Cart3D Config.xml file when running
            comp2tri. If the geometry may be perturbed to a state where one
            component is no longer part of the wetted surface, writing the
            comp2tri file can cause set-up issues, and so it can be beneficial
            to turn it off. The default is True.
        """
        # Define global variable so that functions can access them
        global c3d_logname, _matching_tolerance, _max_matching_tol, _matching_target
        global sens_filename, basefiles_dir, _c3dprepper, sim_dir_name
        global _A_ref, _l_ref
        global _freestream
        global home_dir, working_dir, f_sense_filename
        global obj_cb, jac_cb
        global generator
        global save_comptri, evo_dir
        global moo
        moo = False

        _freestream = freestream

        save_comptri = save_evolution

        generator = vehicle_generator

        # Construct paths
        home_dir = os.getcwd()
        basefiles_dir = os.path.join(home_dir, basefiles_dir_name)
        working_dir = os.path.join(home_dir, working_dir_name)
        sim_dir_name = sim_directory_name
        sens_filename = sensitivity_filename
        f_sense_filename = "F_sensitivities.csv"
        c3d_logname = c3d_log_name

        if save_comptri:
            # Create evolution history directory
            evo_dir = os.path.join(home_dir, "evolution")
            if not os.path.exists(evo_dir):
                os.mkdir(evo_dir)

        # Save callback functions
        obj_cb = objective_callback
        jac_cb = jacobian_callback

        # Save reference area
        _A_ref = A_ref
        _l_ref = l_ref

        # Create instance of Cart3D prepper
        _c3dprepper = C3DPrep(
            logfile=c3d_logname, info_file=c3d_info_file, write_config=write_config_xml
        )

        # Save sensitivity kwargs
        global _sens_kwargs
        _sens_kwargs = sensitivity_kwargs if sensitivity_kwargs else {}

        # Other settings
        _matching_tolerance = matching_tolerance
        _max_matching_tol = 0.1
        _matching_target = 0.9

        # Construct optimisation problem
        self.opt_problem = Optimization(
            name="Cart3D-PySAGAS Shape Optimisation",
            objFun=evaluate_objective,
            sens=evaluate_gradient,
        )

        # Complete super initialisation
        super().__init__(
            optimiser=optimiser,
            working_dir=working_dir,
            optimiser_options=optimiser_options,
        )


class Cart3DMooShapeOpt(Cart3DShapeOpt):
    def __init__(
        self,
        a_inf: float,
        rho_inf: float,
        V_inf: float,
        A_ref: float,
        optimiser: Optimizer,
        vehicle_generator: Generator,
        objective_callback: callable,
        jacobian_callback: callable,
        mach_aoa_points: list[tuple[float, float]],
        optimiser_options: dict = None,
        sensitivity_filename: str = "all_components_sensitivity.csv",
        working_dir_name: str = "working_dir",
        sim_directory_name: str = "simulation",
        basefiles_dir_name: str = "basefiles",
        c3d_log_name: str = "C3D_log",
        c3d_info_file: str = None,
        matching_tolerance: float = 0.00001,
        save_evolution: bool = True,
        write_config_xml: bool = True,
    ) -> None:
        super().__init__(
            a_inf,
            rho_inf,
            V_inf,
            A_ref,
            optimiser,
            vehicle_generator,
            objective_callback,
            jacobian_callback,
            optimiser_options,
            sensitivity_filename,
            working_dir_name,
            sim_directory_name,
            basefiles_dir_name,
            c3d_log_name,
            c3d_info_file,
            matching_tolerance,
            save_evolution,
            write_config_xml,
        )
        global sim_points
        sim_points = mach_aoa_points

        global moo
        moo = True

        # Overload
        self.opt_problem = Optimization(
            name="Cart3D-PySAGAS Shape Optimisation",
            objFun=evaluate_moo_objective,
            sens=evaluate_moo_gradient,
        )

        self.opt_problem: Optimization
        self.optimiser = optimiser(options=optimiser_options)


def evaluate_moo_objective(x: dict) -> dict:
    print("Evaluating objective function.")

    # Pre-process parameters
    _process_parameters(x)

    print("Deploying simulation tasks to pool.")
    pool = mp.Pool()
    results = []
    for result in pool.starmap(aero_wrapper, sim_points):
        results.append(result)
    print("  Done.")

    # Load properties
    properties_dir = glob.glob("*_properties")
    if properties_dir:
        volmass = pd.read_csv(
            glob.glob(os.path.join(properties_dir[0], "*volmass.csv"))[0],
            index_col=0,
        )

        # Fetch user-defined properties
        properties_file = glob.glob(os.path.join(properties_dir[0], "*properties.csv"))
        if len(properties_file) > 0:
            # File exists, load it
            properties = pd.read_csv(
                properties_file[0],
                index_col=0,
            )["0"]
        else:
            properties = None

    # Objective function callback
    funcs = obj_cb(
        parameters=_unwrap_x(x), results=results, volmass=volmass, properties=properties
    )

    return funcs, False


def aero_wrapper(mach, aoa):
    freestream = FlowState(mach=mach, pressure=101e3, temperature=288, aoa=aoa)
    c3d = Cart3D(stl_files=stl_files, freestream=freestream, verbosity=0)
    flow_result = c3d.solve()
    return flow_result


def sens_wrapper(mach, aoa):
    freestream = FlowState(mach=mach, pressure=101e3, temperature=288, aoa=aoa)
    c3d = Cart3D(stl_files=stl_files, freestream=freestream, verbosity=0)

    # Append sensitivity data to tri file
    sim_dir = os.path.join(working_dir, f"M{mach}A{float(aoa)}")
    components_filepath = os.path.join(sim_dir, "Components.i.tri")
    combine_sense_data(
        components_filepath=components_filepath,
        sensitivity_files=sensitivity_files,
        match_target=_matching_target,
        tol_0=_matching_tolerance,
        max_tol=_max_matching_tol,
        outdir=sim_dir,
        verbosity=0,
    )

    # Run the sensitivity solver
    sens_file = os.path.join(sim_dir, "all_components_sensitivity.csv")
    sens_result = c3d.solve_sens(sensitivity_filepath=sens_file)
    flow_result = c3d.flow_result

    return flow_result, sens_result


def evaluate_moo_gradient(x: dict, objective: dict) -> dict:
    """Evaluates the gradient function at the parameter set `x`."""
    print("Evaluating gradient function.")

    # Pre-process parameters
    _process_parameters(x)

    print("Deploying sensitivity tasks to pool.")
    pool = mp.Pool()
    results = []
    for result in pool.starmap(sens_wrapper, sim_points):
        results.append(result)
    print("  Done.")

    # Calculate Jacobian
    properties_dir = glob.glob("*_properties")
    if properties_dir:
        scalar_sens_dir = os.path.join("scalar_sensitivities")
        vm = pd.read_csv(
            glob.glob(os.path.join(properties_dir[0], "*volmass.csv"))[0],
            index_col=0,
        )
        vm_sens = pd.read_csv(
            os.path.join(scalar_sens_dir, "volmass_sensitivity.csv"),
            index_col=0,
        )[x.keys()]

        # Fetch user-defined properties and sensitivities
        properties_file = glob.glob(os.path.join(properties_dir[0], "*properties.csv"))
        if len(properties_file) > 0:
            # File exists, load it
            properties = pd.read_csv(
                properties_file[0],
                index_col=0,
            )["0"]

            # Also load sensitivity file
            property_sens = pd.read_csv(
                os.path.join(scalar_sens_dir, "property_sensitivity.csv"),
                index_col=0,
            )[x.keys()]
        else:
            properties = None
            property_sens = None

    else:
        # No properties data found
        vm = None
        vm_sens = None
        properties = None
        property_sens = None

    # Call function
    jac = jac_cb(
        parameters=_unwrap_x(x),
        results=results,
        volmass=vm,
        volmass_sens=vm_sens,
        properties=properties,
        property_sens=property_sens,
    )

    return jac


def evaluate_objective(x: dict) -> dict:
    """Evaluates the objective function at the parameter set `x`."""
    # Pre-process parameters
    _process_parameters(x)

    # Run Cart3D simulation
    sim_success, loads_dict, _ = _run_simulation()

    # Load properties
    properties_dir = glob.glob("*_properties")
    if properties_dir:
        volmass = pd.read_csv(
            glob.glob(os.path.join(properties_dir[0], "*volmass.csv"))[0],
            index_col=0,
        )

        # Load COG data
        cog = np.loadtxt(
            glob.glob(os.path.join(properties_dir[0], "*cog.txt"))[0], delimiter=","
        )

        # Fetch user-defined properties
        properties_file = glob.glob(os.path.join(properties_dir[0], "*properties.csv"))
        if len(properties_file) > 0:
            # File exists, load it
            properties = pd.read_csv(
                properties_file[0],
                index_col=0,
            )["0"]
        else:
            properties = None

    else:
        # No properties data found
        volmass = None
        properties = None
        cog = None

    if sim_success:
        # Evaluate objective function
        funcs = obj_cb(
            loads_dict=loads_dict,
            volmass=volmass,
            properties=properties,
            cog=cog,
        )
        failed = False

    else:
        # Simulation failed
        funcs = {}
        loads_dict = {"C_L-entire": 0, "C_D-entire": 1}
        funcs = obj_cb(
            loads_dict=loads_dict,
            volmass=volmass,
            properties=properties,
            cog=cog,
        )
        failed = True

    return funcs, failed


def evaluate_gradient(x: dict, objective: dict) -> dict:
    """Evaluates the gradient function at the parameter set `x`."""
    # Pre-process parameters
    _process_parameters(x)

    # Run Cart3D simulation
    _, loads_dict, components_plt_filepath = _run_simulation()

    # Initialise filepaths
    components_filepath = components_plt_filepath
    sensitivity_filepath = os.path.join(working_dir, sens_filename)

    # Create PySAGAS wrapper and run
    try:
        wrapper = Cart3DSensitivityCalculator(
            freestream=_freestream,
            sensitivity_filepath=sensitivity_filepath,
            components_filepath=components_filepath,
            verbosity=0,
        )

    except ValueError:
        # The sensitivity data does not match the point data, regenerate it
        tri_components_filepath = os.path.join(
            working_dir, sim_dir_name, "Components.i.tri"
        )
        sensitivity_files = glob.glob(os.path.join("*sensitivity*"))
        combine_sense_data(
            tri_components_filepath,
            sensitivity_files=sensitivity_files,
            match_target=_matching_target,
            tol_0=_matching_tolerance,
            max_tol=_max_matching_tol,
            outdir=working_dir,
            verbosity=0,
        )

        # Re-instantiate the wrapper
        wrapper = Cart3DSensitivityCalculator(
            freestream=_freestream,
            sensitivity_filepath=sensitivity_filepath,
            components_filepath=components_filepath,
            verbosity=0,
        )

    result = wrapper.calculate(**_sens_kwargs)
    F_sense = result.f_sens
    M_sens = result.m_sens

    # Non-dimensionalise
    coef_sens = F_sense / (_freestream.q * _A_ref)
    moment_coef_sens = M_sens / (_freestream.q * _A_ref * _l_ref)

    # Calculate Jacobian
    properties_dir = glob.glob("*_properties")
    if properties_dir:
        scalar_sens_dir = os.path.join("scalar_sensitivities")
        vm = pd.read_csv(
            glob.glob(os.path.join(properties_dir[0], "*volmass.csv"))[0],
            index_col=0,
        )
        vm_sens = pd.read_csv(
            os.path.join(scalar_sens_dir, "volmass_sensitivity.csv"),
            index_col=0,
        )[x.keys()]

        # Load COG data
        cog = np.loadtxt(
            glob.glob(os.path.join(properties_dir[0], "*cog.txt"))[0], delimiter=","
        )
        cog_sens_files = glob.glob(
            os.path.join(scalar_sens_dir, "*cog_sensitivity.txt")
        )
        cog_sens = {}
        for file in cog_sens_files:
            param = file.split("_cog_sensitivity.txt")[0].split("/")[-1]
            cog_sens[param] = np.loadtxt(file, delimiter=",")

        # Fetch user-defined properties and sensitivities
        properties_file = glob.glob(os.path.join(properties_dir[0], "*properties.csv"))
        if len(properties_file) > 0:
            # File exists, load it
            properties = pd.read_csv(
                properties_file[0],
                index_col=0,
            )["0"]

            # Also load sensitivity file
            property_sens = pd.read_csv(
                os.path.join(scalar_sens_dir, "property_sensitivity.csv"),
                index_col=0,
            )[x.keys()]
        else:
            properties = None
            property_sens = None

    else:
        # No properties data found
        vm = None
        vm_sens = None
        properties = None
        property_sens = None
        cog = None
        cog_sens = None

    # Call function
    failed = False
    try:
        jac = jac_cb(
            parameters=x,
            coef_sens=coef_sens,
            moment_coef_sens=moment_coef_sens,
            loads_dict=loads_dict,
            volmass=vm,
            volmass_sens=vm_sens,
            properties=properties,
            property_sens=property_sens,
            cog=cog,
            cog_sens=cog_sens,
        )

    except Exception as e:
        print(f"Exception calling jacobian callback: {e}")
        jac = {}
        failed = True

    return jac, failed


def _process_parameters(x):
    # Move into working directory
    os.chdir(working_dir)

    # Load existing parameters to compare
    already_started = _compare_parameters(x)

    if not already_started:
        # These parameters haven't run yet, prepare the working directory
        if save_comptri:
            # Move components file into evolution directory
            comp_filepath = os.path.join(working_dir, sim_dir_name, "Components.i.tri")

            # Check if file exists
            if os.path.exists(comp_filepath):
                # Determine new name
                new_filename = f"{len(os.listdir(evo_dir)):04d}.tri"

                # Copy file over
                shutil.copyfile(
                    comp_filepath,
                    os.path.join(evo_dir, new_filename),
                )

        # Delete any files from previous run
        _clean_dir(working_dir)

    # Dump design parameters to file
    with open("parameters.pkl", "wb") as f:
        pickle.dump(x, f)

    # Generate vehicle and geometry sensitivities
    if len(glob.glob("*sensitivity*")) == 0 or not already_started:
        # No sensitivity files generated yet, or this is new geometry
        print("Running sensitivity study.")
        parameters = _unwrap_x(x)
        ss = SensitivityStudy(vehicle_constructor=generator, verbosity=0)
        ss.dvdp(parameter_dict=parameters, perturbation=2, write_nominal_stl=True)
        ss.to_csv()
        print("  Done.")

    if moo:
        # Multi-objective optimisation
        global stl_files, sensitivity_files
        stl_files = glob.glob("*.stl")
        sensitivity_files = glob.glob("*sensitivity*")

        # Also copy aero.csh and input.cntl into working directory
        for filename in ["input.cntl", "aero.csh"]:
            shutil.copyfile(
                os.path.join(home_dir, filename),
                os.path.join(working_dir, filename),
            )
            shutil.copymode(
                os.path.join(home_dir, filename),
                os.path.join(working_dir, filename),
            )


def _run_simulation(no_attempts: int = 3):
    """Prepare and run the CFD simulation with Cart3D. The simulation will be
    run in the 'simulation' subdirectory of the iteration directory.
    """
    # Make simulation directory
    sim_dir = os.path.join(working_dir, sim_dir_name)
    run_intersect = False
    components_filepath = os.path.join(sim_dir, "Components.i.tri")
    if not os.path.exists(sim_dir):
        os.mkdir(sim_dir)
        run_intersect = True
    else:
        # Check for intersected file
        run_intersect = not os.path.exists(components_filepath)
        intersected = True

    # Attempt component intersection
    sim_success = False
    for attempt in range(no_attempts):
        # Run intersect
        if run_intersect:
            _c3dprepper._log(f"SHAPEOPT INTERSECT ATTEMPT {attempt+1}")
            intersected = _c3dprepper.intersect_stls()

        # Check for intersection
        if intersected:
            # Prepare rest of simulation directory
            if not os.path.exists(os.path.join(sim_dir, "input.cntl")):
                # Move files to simulation directory (including Components.i.tri)
                _c3dprepper.run_autoinputs()
                os.system(
                    f"mv *.tri Config.xml input.c3d preSpec.c3d.cntl {sim_dir} >> {c3d_logname} 2>&1"
                )

                # Copy sim files and permissions
                for filename in ["input.cntl", "aero.csh"]:
                    shutil.copyfile(
                        os.path.join(basefiles_dir, filename),
                        os.path.join(sim_dir, filename),
                    )
                    shutil.copymode(
                        os.path.join(basefiles_dir, filename),
                        os.path.join(sim_dir, filename),
                    )

            # Create all_components_sensitivity.csv
            if not os.path.exists(sens_filename):
                combine_sense_data(
                    components_filepath,
                    sensitivity_files=glob.glob("*sensitivity*"),
                    match_target=_matching_target,
                    tol_0=_matching_tolerance,
                    max_tol=_max_matching_tol,
                    verbosity=0,
                )

            # Run Cart3D and await result
            os.chdir(sim_dir)
            target_adapt = _infer_adapt(sim_dir)
            c3d_donefile = os.path.join(sim_dir, target_adapt, "FLOW", "DONE")
            run_cmd = "./aero.csh restart"
            _restarts = 0
            if not os.path.exists(c3d_donefile):
                with open(c3d_logname, "a") as f:
                    subprocess.run(
                        run_cmd, shell=True, stdout=f, stderr=subprocess.STDOUT
                    )
                while not os.path.exists(c3d_donefile):
                    # Wait...
                    time.sleep(5)

                    # Check for C3D failure
                    running, e = _c3d_running()

                    if not running:
                        # C3D failed, try restart it
                        if _restarts > 3:
                            return False

                        f = open(c3d_logname, "a")
                        subprocess.run(
                            run_cmd, shell=True, stdout=f, stderr=subprocess.STDOUT
                        )
                        f.close()
                        _restarts += 1

            sim_success = True
            break

    if sim_success:
        # Sim finished successfully, read loads file
        loads_dict = _read_c3d_loads(
            os.path.join(working_dir, sim_dir_name, "BEST/FLOW/loadsCC.dat")
        )
        components_plt_filepath = os.path.join(
            working_dir, sim_dir_name, "BEST/FLOW/Components.i.plt"
        )

    else:
        loads_dict = None
        components_plt_filepath = None

    # Change back to working directory
    os.chdir(working_dir)

    return sim_success, loads_dict, components_plt_filepath


def _read_c3d_loads(
    loadsCC_filepath: str,
    b_frame: bool = True,
    v_frame: bool = True,
    moments: bool = True,
) -> dict:
    load_dict = {}
    with open(loadsCC_filepath, "r") as file:
        for line in file:
            if line[0] == "#":
                # Commented line, skip
                continue

            # Remove linebreaks and multiple spaces
            line = " ".join(line.split())
            words = line.split(":")

            if len(words) == 0 or len(words) == 1:  # skip if empty line
                continue

            text = words[0]
            number = float(words[1])
            word = text.split(" ")
            tag = word[0]
            coeff = word[-1]
            coeff = coeff[1:4]
            if b_frame is True:
                if coeff in ["C_A", "C_Y", "C_N"]:  # get force in b_frame
                    load_dict["{0}-{1}".format(coeff, tag)] = number
            if v_frame is True:
                if coeff in ["C_D", "C_S", "C_L"]:  # get force in v_frame
                    load_dict["{0}-{1}".format(coeff, tag)] = number
            if moments is True:
                if coeff in ["C_l", "C_m", "C_n", "C_M"]:  # get moment coeff
                    load_dict["{0}-{1}".format(coeff, tag)] = number

    return load_dict


def _c3d_running() -> bool:
    """Watches the Cart3D log file to check for errors and return False
    if Cart3D has stopped running."""
    with open(c3d_logname) as f:
        # Get last line in log file
        for line in f:
            pass

        # Check if if it is in the known errors
        for e in Cart3DShapeOpt.C3D_errors:
            if e in line:
                return False, e

    # No errors
    return True, None


def _infer_adapt(sim_dir) -> str:
    with open(f"{sim_dir}/aero.csh", "r") as f:
        lines = f.readlines()

        for line in lines:
            if line.find("set n_adapt_cycles") != -1:
                return f"adapt{int(line.split('=')[-1]):02d}"


def _compare_parameters(x):
    """Compares the current parameters x to the last run simulation
    parameters."""
    try:
        with open("parameters.pkl", "rb") as f:
            xp = pickle.load(f)

        # Compare to current parameters
        already_run = x == xp

    except FileNotFoundError:
        # Simulation not run yet
        already_run = False

    return already_run


def _clean_dir(directory: str, keep: list = None):
    """Deletes everything in a directory except for what is specified
    in keep."""
    all_files = os.listdir(directory)
    if keep is None:
        # Convert to empty list
        keep = []
    rm_files = set(all_files) - set(keep)
    for f in rm_files:
        if os.path.isdir(f):
            shutil.rmtree(f)
        else:
            os.remove(f)


def _unwrap_x(x: dict) -> dict:
    """Unwraps an ordered dictionary."""
    unwrapped = {}
    for key, val in x.items():
        if len(val) == 1:
            unwrapped[key] = val[0]
        else:
            unwrapped[key] = val
    return unwrapped



import os
import sys
import traceback
import pandas as pd
from typing import Tuple, List


def process_components_file(
    a_inf: float,
    rho_inf: float,
    filepath: str = "Components.i.plt",
    write_data: bool = True,
    verbosity: int = 1,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """A ParaView script to process Components.i.plt to extract
    points and cells with data attached.

    Parameters
    ----------
    a_inf : float
        The freestream speed of sound (m/s).

    rho_inf : float
        The freestream density (kg/m^3).

    filepath : str, optional
        The filepath to the Components.i.plt file to be processed.
        The default is Components.i.plt.

    write_data : bool, optional
        Write the flow data to CSV files. The default is True.

    verbosity : int, optional
        The verbosity of the code. The defualt is 1.

    Returns
    -------
    points : pd.DataFrame
        A DataFrame of the point data.
    cells : pd.DataFrame
        A DataFrame of the cell data.
    """
    # Check file exists
    if not os.path.exists(filepath):
        raise FileNotFoundError(f"File '{filepath}' does not exist.")

    # Define root directory
    root_dir = os.path.dirname(filepath)

    try:
        # import the simple module from the paraview
        import vtk
        from paraview.simple import (
            VisItTecplotBinaryReader,
            _DisableFirstRenderCameraReset,
            GetActiveViewOrCreate,
            Show,
            ProgrammableFilter,
            Hide,
            PointDatatoCellData,
            ExportView,
        )
    except ModuleNotFoundError:
        # Cannot find paraview python package
        print(
            "Cannot find ParaView Python package. If ParaView is already "
            + "installed, please append the bin/ directory to the Python path. "
            + "If it is not installed, please do so. If you are using "
            + "an Anaconda environment, you can install using "
            + "'conda install -c conda-forge paraview'.\n\n"
        )
        tb = traceback.format_exc()
        print(tb)
        sys.exit()

    # disable automatic camera reset on 'Show'
    _DisableFirstRenderCameraReset()

    # create a new 'VisItTecplotBinaryReader'
    if verbosity > 0:
        print(f"Loading {filepath}")
    componentsiplt = VisItTecplotBinaryReader(FileName=[filepath])
    componentsiplt.MeshStatus = ["Surface"]
    componentsiplt.PointArrayStatus = []

    # Properties modified on componentsiplt
    componentsiplt.PointArrayStatus = [
        "Cp",
        "Pressure",
        "Rho",
        "U",
        "V",
        "W",
        "x",
        "y",
        "z",
    ]

    # get active view
    spreadSheetView1 = GetActiveViewOrCreate("SpreadSheetView")
    # uncomment following to set a specific view size
    # spreadSheetView1.ViewSize = [400, 400]

    # show data in view
    componentsipltDisplay = Show(componentsiplt, spreadSheetView1)

    # update the view to ensure updated data information
    spreadSheetView1.Update()

    # create a new 'Programmable Filter'
    programmableFilter1 = ProgrammableFilter(Input=componentsiplt)
    programmableFilter1.Script = ""
    programmableFilter1.RequestInformationScript = ""
    programmableFilter1.RequestUpdateExtentScript = ""
    programmableFilter1.PythonPath = ""

    # Properties modified on programmableFilter1
    if verbosity > 0:
        print("  Dimensionalising attributes.")
    programmableFilter1.Script = f"""
    # Paraview script to normalise Cart3D variables.
    R_gas = 287.058
    gamma = 1.4

    # Redefine the Cart3D variables - normalised values are *_tilde
    output.PointData.append(inputs[0].PointData["U"], "U_tilde")
    output.PointData.append(inputs[0].PointData["V"], "V_tilde")
    output.PointData.append(inputs[0].PointData["W"], "W_tilde")
    output.PointData.append(inputs[0].PointData["Pressure"], "p_tilde")
    output.PointData.append(inputs[0].PointData["Rho"], "rho_tilde")

    # Define the dimensional flow properties
    output.PointData.append({a_inf} * inputs[0].PointData["U"], "U")
    output.PointData.append({a_inf} * inputs[0].PointData["V"], "V")
    output.PointData.append({a_inf} * inputs[0].PointData["W"], "W")

    output.PointData.append({a_inf} * {a_inf} * {rho_inf} * output.PointData["p_tilde"], "p")
    output.PointData.append({rho_inf} * output.PointData["rho_tilde"], "rho")
    output.PointData.append({a_inf}*{a_inf}*{rho_inf}*output.PointData["p_tilde"] / ({rho_inf}*output.PointData["rho_tilde"]*R_gas), "T")
    output.PointData.append((abs(gamma * R_gas * output.PointData["T"]))**0.5, "a")
    output.PointData.append((output.PointData["U"]**2 + output.PointData["V"]**2 + output.PointData["W"]**2)**0.5, "V_mag")
    output.PointData.append(output.PointData["V_mag"]/output.PointData["a"], "M")"""
    programmableFilter1.RequestInformationScript = ""
    programmableFilter1.RequestUpdateExtentScript = ""
    programmableFilter1.PythonPath = ""

    # show data in view
    programmableFilter1Display = Show(programmableFilter1, spreadSheetView1)

    # hide data in view
    Hide(componentsiplt, spreadSheetView1)

    # update the view to ensure updated data information
    spreadSheetView1.Update()

    # create a new 'Point Data to Cell Data'
    pointDatatoCellData1 = PointDatatoCellData(Input=programmableFilter1)

    # show data in view
    pointDatatoCellData1Display = Show(pointDatatoCellData1, spreadSheetView1)

    # hide data in view
    Hide(programmableFilter1, spreadSheetView1)

    # update the view to ensure updated data information
    spreadSheetView1.Update()

    # Properties modified on pointDatatoCellData1
    pointDatatoCellData1.PassPointData = 1

    # update the view to ensure updated data information
    spreadSheetView1.Update()

    # Properties modified on spreadSheetView1
    spreadSheetView1.GenerateCellConnectivity = 1

    # update the view to ensure updated data information
    spreadSheetView1.Update()

    # Properties modified on spreadSheetView1
    spreadSheetView1.FieldAssociation = "Point Data"

    # update the view to ensure updated data information
    spreadSheetView1.Update()

    # export view
    if verbosity > 0:
        print("  Saving point data.")
    points_filename = os.path.join(root_dir, "points.csv")
    ExportView(points_filename, view=spreadSheetView1)

    # Properties modified on spreadSheetView1
    spreadSheetView1.FieldAssociation = "Cell Data"

    # update the view to ensure updated data information
    spreadSheetView1.Update()

    # export view
    if verbosity > 0:
        print("  Saving cell data.")
    cells_filename = os.path.join(root_dir, "cells.csv")
    ExportView(cells_filename, view=spreadSheetView1)

    if verbosity > 0:
        print("Complete.")

    # Read the files to Pandas DataFrames
    points = pd.read_csv(points_filename)
    cells = pd.read_csv(cells_filename)

    if not write_data:
        # Delete the files
        os.remove(points_filename)
        os.remove(cells_filename)

    return points, cells
